from mpi4py import MPI
import dolfinx
from scifem import create_space_of_simple_functions, assemble_scalar
import pytest
import numpy as np
import ufl


@pytest.mark.parametrize("value_shape", [(), (3,), (3, 2)])
@pytest.mark.parametrize(
    "cell_type", ["interval", "triangle", "quadrilateral", "hexahedron", "tetrahedron"]
)
def test_material_space(value_shape: tuple[int], cell_type: str):
    cell = dolfinx.mesh.to_type(cell_type)
    tdim = dolfinx.mesh.cell_dim(cell)
    if tdim == 1:
        mesh = dolfinx.mesh.create_unit_interval(
            MPI.COMM_WORLD, 20, ghost_mode=dolfinx.mesh.GhostMode.shared_facet
        )
    elif tdim == 2:
        mesh = dolfinx.mesh.create_unit_square(
            MPI.COMM_WORLD, 20, 20, ghost_mode=dolfinx.mesh.GhostMode.shared_facet, cell_type=cell
        )
    elif tdim == 3:
        mesh = dolfinx.mesh.create_unit_cube(
            MPI.COMM_WORLD,
            12,
            12,
            8,
            ghost_mode=dolfinx.mesh.GhostMode.shared_facet,
            cell_type=cell,
        )
    else:
        raise ValueError(f"Unsupported {cell_type=}")

    def ball(x, tol=1e-8):
        conditions = []
        for i in range(tdim):
            conditions.append((x[i] - 0.5) ** 2 <= 0.2**2 + tol)
        return np.logical_and.reduce(conditions)

    def top(x, tol=1e-8):
        return x[tdim - 1] >= 0.7 - tol

    cell_map = mesh.topology.index_map(mesh.topology.dim)
    all_cells = np.arange(cell_map.size_local + cell_map.num_ghosts, dtype=np.int32)
    tags = [1, 5, 3]
    values = np.full_like(all_cells, tags[0])
    values[dolfinx.mesh.locate_entities(mesh, mesh.topology.dim, ball)] = tags[1]
    values[dolfinx.mesh.locate_entities(mesh, mesh.topology.dim, top)] = tags[2]
    ct = dolfinx.mesh.meshtags(mesh, mesh.topology.dim, all_cells, values)

    V = create_space_of_simple_functions(mesh, ct, tags, value_shape=value_shape)
    assert V.dofmap.index_map.size_global == len(tags)
    assert V.dofmap.index_map.size_local + V.dofmap.index_map.num_ghosts == len(tags)
    assert (V.dofmap.index_map.size_local == 0) or (V.dofmap.index_map.num_ghosts == 0)
    value_size = int(np.prod(value_shape))
    assert V.dofmap.index_map_bs == value_size
    assert V.dofmap.bs == value_size
    array = V.dofmap.list
    for i, tag in enumerate(tags):
        np.testing.assert_allclose(array[ct.indices[ct.values == tag]].flatten(), i)

    # Check that we can assign values to the function for each region, and that it gives
    # a sensible result
    bs = V.dofmap.bs
    for i, tag in enumerate(tags):
        vals = np.arange(i + 1, i + 1 + bs, dtype=np.float64)
        f = dolfinx.fem.Function(V)
        f.x.array[i * bs : (i + 1) * bs] = vals

        form = dolfinx.fem.form(ufl.inner(f, f) * ufl.dx)
        integral = assemble_scalar(form)

        # Compute integral of exact function on the restricted domain
        dx_restricted = ufl.Measure("dx", domain=mesh, subdomain_data=ct, subdomain_id=tag)
        if value_shape == ():
            f_ex = vals[0]
        else:
            f_ex = ufl.as_tensor(vals.reshape(value_shape))
        form_ex = ufl.inner(f_ex, f_ex) * dx_restricted
        integral_ex = assemble_scalar(form_ex)
        np.testing.assert_allclose(integral, integral_ex)

        # Compute squared error on the restricted domain
        diff = ufl.inner(f - f_ex, f - f_ex) * dx_restricted
        error = assemble_scalar(diff)
        np.testing.assert_allclose(error, 0.0)



from mpi4py import MPI
from petsc4py import PETSc
import dolfinx
import dolfinx.fem.petsc
import ufl
import numpy as np
import scifem
import pytest


@pytest.mark.parametrize("L", [0.1, 0.2, 0.3])
@pytest.mark.parametrize("H", [1.3, 0.8, 0.2])
@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
@pytest.mark.parametrize("dtype", [np.float64, np.float32])
def test_real_function_space_mass(L, H, cell_type, dtype):
    """
    Check that real space mass matrix is the same as assembling the volume of the mesh
    """

    mesh = dolfinx.mesh.create_rectangle(
        MPI.COMM_WORLD, [[0.0, 0.0], [L, H]], [7, 9], cell_type, dtype=dtype
    )

    V = scifem.create_real_functionspace(mesh)
    u = ufl.TrialFunction(V)
    v = ufl.TestFunction(V)
    a = ufl.inner(u, v) * ufl.dx

    A = dolfinx.fem.assemble_matrix(dolfinx.fem.form(a, dtype=dtype), bcs=[])
    A.scatter_reverse()
    tol = 100 * np.finfo(dtype).eps
    cell_map = mesh.topology.index_map(mesh.topology.dim)
    if cell_map.size_local + cell_map.num_ghosts > 0:
        assert len(A.data) == 1
        if cell_map.local_range[0] == 0:
            assert np.isclose(A.data[0], L * H, atol=tol)
    else:
        assert len(A.data) == 0
        assert len(V.dofmap.list.flatten()) == 0
        assert V.dofmap.index_map.size_local == 0
        assert V.dofmap.index_map.num_ghosts == 1


@pytest.mark.parametrize("dtype", [np.float64, np.float32])
@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.tetrahedron, dolfinx.mesh.CellType.hexahedron]
)
def test_real_function_space_vector(cell_type, dtype):
    """
    Test that assembling against a real space test function is equivalent to assembling a vector
    """

    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 2, 3, 5, cell_type, dtype=dtype)

    V = dolfinx.fem.functionspace(mesh, ("Lagrange", 3))
    v = ufl.TrialFunction(V)

    R = scifem.create_real_functionspace(mesh)
    u = ufl.TestFunction(R)
    a_R = ufl.inner(u, v) * ufl.dx
    form_rhs = dolfinx.fem.form(a_R, dtype=dtype)

    A_R = dolfinx.fem.assemble_matrix(form_rhs, bcs=[])
    A_R.scatter_reverse()

    L = ufl.inner(ufl.constantvalue.IntValue(1), v) * ufl.dx
    form_lhs = dolfinx.fem.form(L, dtype=dtype)
    b = dolfinx.fem.assemble_vector(form_lhs)
    b.scatter_reverse(dolfinx.la.InsertMode.add)
    b.scatter_forward()

    row_map = A_R.index_map(0)
    num_local_rows = row_map.size_local
    num_dofs = V.dofmap.index_map.size_local * V.dofmap.index_map_bs
    tol = 100 * np.finfo(dtype).eps
    if MPI.COMM_WORLD.rank == 0:
        assert num_local_rows == 1
        num_dofs_global = V.dofmap.index_map.size_global * V.dofmap.index_map_bs
        assert A_R.indptr[1] - A_R.indptr[0] == num_dofs_global
        np.testing.assert_allclose(A_R.indices, np.arange(num_dofs_global))
        np.testing.assert_allclose(b.array[:num_dofs], A_R.data[:num_dofs], atol=tol)
    else:
        assert num_local_rows == 0


@pytest.mark.parametrize("dtype", [(PETSc.RealType, PETSc.ScalarType)])
@pytest.mark.parametrize("tensor", [0, 1, 2])
@pytest.mark.parametrize("degree", range(1, 5))
def test_singular_poisson(tensor, degree, dtype):
    M = 9

    rtype, stype = dtype
    mesh = dolfinx.mesh.create_unit_square(
        MPI.COMM_WORLD, M, M, dolfinx.mesh.CellType.triangle, dtype=rtype
    )

    if tensor == 0:
        value_shape = ()
    elif tensor == 1:
        value_shape = (2,)
    else:
        value_shape = (3, 2)

    V = dolfinx.fem.functionspace(mesh, ("Lagrange", degree, value_shape))
    R = scifem.create_real_functionspace(mesh, value_shape)

    u = ufl.TrialFunction(V)
    v = ufl.TestFunction(V)
    c = ufl.TrialFunction(R)
    d = ufl.TestFunction(R)
    x = ufl.SpatialCoordinate(mesh)
    pol = x[0] ** degree - 2 * x[1] ** degree
    # Compute average value of polynomial to make mean 0
    C = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(pol * ufl.dx, dtype=stype)), op=MPI.SUM
    )
    u_scalar = pol - dolfinx.fem.Constant(mesh, stype(C))
    if tensor == 0:
        u_ex = u_scalar
        zero = dolfinx.fem.Constant(mesh, stype(0.0))
    elif tensor == 1:
        u_ex = ufl.as_vector([u_scalar, -u_scalar])
        zero = dolfinx.fem.Constant(mesh, stype((0.0, 0.0)))
    else:
        u_ex = ufl.as_tensor(
            [
                [u_scalar, 2 * u_scalar],
                [3 * u_scalar, -u_scalar],
                [u_scalar, 2 * u_scalar],
            ]
        )
        zero = dolfinx.fem.Constant(mesh, stype(((0.0, 0.0), (0.0, 0.0), (0.0, 0.0))))

    dx = ufl.Measure("dx", domain=mesh)
    f = -ufl.div(ufl.grad(u_ex))
    n = ufl.FacetNormal(mesh)
    g = ufl.dot(ufl.grad(u_ex), n)
    a00 = ufl.inner(ufl.grad(u), ufl.grad(v)) * dx
    a01 = ufl.inner(c, v) * dx
    a10 = ufl.inner(u, d) * dx

    L0 = ufl.inner(f, v) * dx + ufl.inner(g, v) * ufl.ds
    L1 = ufl.inner(zero, d) * dx

    a = dolfinx.fem.form([[a00, a01], [a10, None]], dtype=stype)
    L = dolfinx.fem.form([L0, L1], dtype=stype)

    new_assemble_mode = False
    try:
        A = dolfinx.fem.petsc.assemble_matrix_block(a)
    except AttributeError:
        new_assemble_mode = True
        A = dolfinx.fem.petsc.assemble_matrix(a, kind="mpi")
    A.assemble()

    if new_assemble_mode:
        b = dolfinx.fem.petsc.create_vector(dolfinx.fem.extract_function_spaces(L), kind="mpi")
    else:
        b = dolfinx.fem.petsc.create_vector_block(L)

    with b.localForm() as loc:
        loc.set(0)

    if new_assemble_mode:
        dolfinx.fem.petsc.assemble_vector(b, L)
        b.ghostUpdate(addv=PETSc.InsertMode.ADD, mode=PETSc.ScatterMode.REVERSE)
        b.ghostUpdate(addv=PETSc.InsertMode.INSERT, mode=PETSc.ScatterMode.FORWARD)
    else:
        dolfinx.fem.petsc.assemble_vector_block(b, L, a, bcs=[])

    ksp = PETSc.KSP().create(mesh.comm)
    ksp.setOperators(A)
    ksp.setType("preonly")
    pc = ksp.getPC()
    pc.setType("lu")
    pc.setFactorSolverType("mumps")

    if new_assemble_mode:
        x = dolfinx.fem.petsc.create_vector(dolfinx.fem.extract_function_spaces(L), kind="mpi")
    else:
        x = dolfinx.fem.petsc.create_vector_block(L)

    ksp.solve(b, x)
    x.ghostUpdate(addv=PETSc.InsertMode.INSERT, mode=PETSc.ScatterMode.FORWARD)

    uh = dolfinx.fem.Function(V)
    rh = dolfinx.fem.Function(R)
    if new_assemble_mode:
        dolfinx.fem.petsc.assign(x, [uh, rh])
    else:
        x_local = dolfinx.cpp.la.petsc.get_local_vectors(
            x,
            [
                (V.dofmap.index_map, V.dofmap.index_map_bs),
                (R.dofmap.index_map, R.dofmap.index_map_bs),
            ],
        )
        uh.x.array[: len(x_local[0])] = x_local[0]
        uh.x.scatter_forward()

    b.destroy()
    x.destroy()
    A.destroy()
    ksp.destroy()

    error = dolfinx.fem.form(ufl.inner(u_ex - uh, u_ex - uh) * dx, dtype=stype)

    e_local = dolfinx.fem.assemble_scalar(error)
    tol = 1e4 * np.finfo(stype).eps
    e_global = np.sqrt(mesh.comm.allreduce(e_local, op=MPI.SUM))
    assert np.isclose(e_global, 0, atol=tol)


@pytest.mark.parametrize("ftype, stype", [(np.float32, np.complex64), (np.float64, np.complex128)])
def test_complex_real_space(ftype, stype):
    mesh = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, 13, dtype=ftype)

    val = (2 + 3j, -4 + 5j)
    value_shape = (2,)
    R = scifem.create_real_functionspace(mesh, value_shape=value_shape)
    u = dolfinx.fem.Function(R, dtype=stype)
    u.x.array[0] = val[0]
    u.x.array[1] = val[1]
    u.x.scatter_forward()

    V = dolfinx.fem.functionspace(mesh, ("Lagrange", 1, value_shape))
    v = ufl.TestFunction(V)
    L = ufl.inner(u, v) * ufl.dx

    b = dolfinx.fem.assemble_vector(dolfinx.fem.form(L, dtype=stype))
    b.scatter_reverse(dolfinx.la.InsertMode.add)
    b.scatter_forward()
    const = dolfinx.fem.Constant(mesh, stype(val))
    L_const = ufl.inner(const, v) * ufl.dx
    b_const = dolfinx.fem.assemble_vector(dolfinx.fem.form(L_const, dtype=stype))
    b_const.scatter_reverse(dolfinx.la.InsertMode.add)
    b_const.scatter_forward()

    tol = 100 * np.finfo(stype).eps
    np.testing.assert_allclose(b.array, b_const.array, atol=tol)



from mpi4py import MPI

import numpy as np
import dolfinx
import dolfinx.nls.petsc
from petsc4py import PETSc
from scifem import assemble_scalar, BlockedNewtonSolver, NewtonSolver
import basix.ufl
import ufl
import pytest


@pytest.mark.parametrize("factor", [1, -1])
@pytest.mark.parametrize("auto_split", [True, False])
def test_NewtonSolver(factor, auto_split):
    def left_boundary(x):
        return np.isclose(x[0], 0.0)

    dtype = PETSc.RealType
    ftype = PETSc.ScalarType
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 12, dtype=dtype)
    mesh.topology.create_connectivity(mesh.topology.dim - 1, mesh.topology.dim)
    boundary_facets = dolfinx.mesh.locate_entities_boundary(mesh, 1, left_boundary)

    el_0 = basix.ufl.element("Lagrange", mesh.basix_cell(), 1)
    el_1 = basix.ufl.element("Lagrange", mesh.basix_cell(), 2)
    V = dolfinx.fem.functionspace(mesh, el_0)
    Q = dolfinx.fem.functionspace(mesh, el_1)

    if auto_split:
        W = ufl.MixedFunctionSpace(V, Q)
        v, q = ufl.TestFunctions(W)
        du, dp = ufl.TrialFunctions(W)
    else:
        v = ufl.TestFunction(V)
        q = ufl.TestFunction(Q)
        du = ufl.TrialFunction(V)
        dp = ufl.TrialFunction(Q)

    # Set nonzero initial guess
    # Choose initial guess acording to the input factor
    u = dolfinx.fem.Function(V, dtype=ftype)
    u.x.array[:] = factor * 0.1
    p = dolfinx.fem.Function(Q, dtype=ftype)
    p.x.array[:] = factor * 0.02
    x = ufl.SpatialCoordinate(mesh)
    u_expr = 3 * x[0] ** 2
    p_expr = 5 * x[1] ** 4
    c0 = dolfinx.fem.Constant(mesh, ftype(0.3))
    c1 = dolfinx.fem.Constant(mesh, ftype(0.82))
    F0 = ufl.inner(c0 * u**2, v) * ufl.dx - ufl.inner(u_expr, v) * ufl.dx
    F1 = ufl.inner(c1 * p**2, q) * ufl.dx - ufl.inner(p_expr, q) * ufl.dx
    if auto_split:
        F_ = F0 + F1
        F = list(ufl.extract_blocks(F_))
        J = ufl.extract_blocks(ufl.derivative(F_, u, du) + ufl.derivative(F_, p, dp))
    else:
        F = [F0, F1]
        J = [
            [ufl.derivative(F0, u, du), ufl.derivative(F0, p, dp)],
            [ufl.derivative(F1, u, du), ufl.derivative(F1, p, dp)],
        ]

    # Reference solution
    u_ex = factor * ufl.sqrt(u_expr / c0)
    p_ex = factor * ufl.sqrt(p_expr / c1)

    # Create BC on second space
    dofs_q = dolfinx.fem.locate_dofs_topological(Q, mesh.topology.dim - 1, boundary_facets)
    p_bc = dolfinx.fem.Function(Q, dtype=ftype)
    try:
        ip = Q.element.interpolation_points()

    except TypeError:
        ip = Q.element.interpolation_points
    p_bc.interpolate(dolfinx.fem.Expression(p_ex, ip))
    bc_q = dolfinx.fem.dirichletbc(p_bc, dofs_q)

    petsc_options = {"ksp_type": "preonly", "pc_type": "lu", "pc_factor_mat_solver_type": "mumps"}
    solver = NewtonSolver(F, J, [u, p], bcs=[bc_q], max_iterations=25, petsc_options=petsc_options)
    solver.solve()

    err_u = ufl.inner(u_ex - u, u_ex - u) * ufl.dx
    err_p = ufl.inner(p_ex - p, p_ex - p) * ufl.dx
    tol = np.finfo(dtype).eps * 1.0e3
    assert assemble_scalar(err_u) < tol
    assert assemble_scalar(err_p) < tol

    # Check consistency with other Newton solver
    blocked_solver = BlockedNewtonSolver(
        F, [u, p], bcs=[bc_q], J=None if auto_split else J, petsc_options=petsc_options
    )

    dolfinx.log.set_log_level(dolfinx.log.LogLevel.ERROR)
    u.x.array[:] = factor * 0.1
    p.x.array[:] = factor * 0.02
    blocked_solver.convergence_criterion = "incremental"
    dolfinx.log.set_log_level(dolfinx.log.LogLevel.INFO)
    blocked_solver.solve()

    err_u = ufl.inner(u_ex - u, u_ex - u) * ufl.dx
    err_p = ufl.inner(p_ex - p, p_ex - p) * ufl.dx
    tol = np.finfo(dtype).eps * 1.0e3
    assert assemble_scalar(err_u) < tol
    assert assemble_scalar(err_p) < tol



from packaging.version import Version
import dolfinx
import ufl
import numpy as np
import numpy.typing as npt
from .utils import unroll_dofmap

__all__ = ["interpolation_matrix", "prepare_interpolation_data", "interpolate_to_surface_submesh"]

if dolfinx.has_petsc4py:
    from petsc4py import PETSc

    __all__.append("petsc_interpolation_matrix")


def prepare_interpolation_data(
    expr: ufl.core.expr.Expr,
    Q: dolfinx.fem.FunctionSpace,
    interpolation_entities: npt.NDArray[np.int32] | None = None,
) -> npt.NDArray[np.inexact]:
    """Convenience function for preparing data required for assembling the interpolation matrix

    .. math::
        \\begin{align*}
        \\Lambda: V &\\rightarrow Q \\\\
        \\Lambda u &= \\sum_{i=0}^{N_Q-1}\\sum_{j=0}^{N_V-1} \\phi_i l_i(expr(\\psi_j))u_j
        \\end{align*}

    where :math:`l_j` is the dual basis of the space :math:`Q` with basis functions :math:`\\phi_j`,
    and :math:`\\psi_j` are the basis functions of the space :math:`V`.

    Args:
        expr: The UFL expression containing a trial function from space `V`
        Q: Output interpolation space
        interpolation_entities: Entities of the domain of the input space `V` that one
            should evaluate the `expr` at. If not provided, it is assumed that
            we are integrating over all cells in `V` and that `Q` is defined on the same grid.
    Returns:
        Interpolation data per cell, as an numpy array.
    """
    if np.issubdtype(dolfinx.default_scalar_type, np.complexfloating):
        raise NotImplementedError("No complex support")

    # Extract argument from expr (in V)
    arguments = ufl.algorithms.extract_arguments(expr)
    assert len(arguments) == 1
    V = arguments[0].ufl_function_space()

    mesh = V.mesh

    if Q.mesh.topology.dim == V.mesh.topology.dim:
        if interpolation_entities is None:
            tdim = mesh.topology.dim
            num_cells = mesh.topology.index_map(tdim).size_local
            interpolation_entities = np.arange(num_cells, dtype=np.int32)
        else:
            if (ndim := interpolation_entities.ndim) != 1:
                raise ValueError(
                    f"Interpolation entities has wrong input shape, should be 1D, got {ndim}"
                )
            num_cells = len(interpolation_entities)
    elif Q.mesh.topology.dim == V.mesh.topology.dim - 1:
        if interpolation_entities is None:
            raise ValueError(
                "For integration onto a submesh of codim 1,"
                + "the integration entities has to be provided"
            )
        else:
            if (ndim := interpolation_entities.ndim) != 2:
                raise ValueError(
                    f"Interpolation entities has wrong input shape, should be 2D, got {ndim}"
                )
            num_cells = interpolation_entities.shape[0]
    else:
        raise RuntimeError("Only codim-1 interpolation matrices can be defined")

    # Extract quadrature points for expression (in Q space)
    try:
        q_points = Q.element.interpolation_points()
    except TypeError:
        q_points = Q.element.interpolation_points

    # Compile expression
    num_points = q_points.shape[0]
    compiled_expr = dolfinx.fem.Expression(expr, q_points)

    # (num_cells, num_points, num_dofs*bs, expr_value_size)
    array_evaluated = compiled_expr.eval(mesh, interpolation_entities)
    assert np.prod(Q.value_shape) == np.prod(expr.ufl_shape)

    # Get data as (num_cells*num_points,1, expr_shape, num_test_basis_functions*test_block_size)
    expr_size = int(np.prod(expr.ufl_shape))
    array_evaluated = array_evaluated.reshape(
        num_cells * q_points.shape[0], 1, expr_size, V.dofmap.bs * V.dofmap.dof_layout.num_dofs
    )

    # Check if we are dealing with a quadrature element or not.
    # They do not have a complete DOLFINx API, which makes them tricky to use.
    try:
        basix_el = Q.element.basix_element
        Q_vs = basix_el.value_size
        pull_back = basix_el.pull_back
        im = basix_el.interpolation_matrix
    except RuntimeError:
        Q_vs = 1  # If we do not have a basix element, assume value size is 1
        assert isinstance(Q.ufl_element().pullback, ufl.pullback.IdentityPullback)
        pull_back = lambda x: None
        assert Q.element.interpolation_ident
        im = None

    new_array = np.zeros(
        (num_cells * num_points, Q.dofmap.bs * Q_vs, V.dofmap.bs * V.dofmap.dof_layout.num_dofs),
        dtype=np.float64,
    )

    # Check if pullback is identity, then we can skip this step
    if not isinstance(Q.ufl_element().pullback, ufl.pullback.IdentityPullback):
        jacobian = dolfinx.fem.Expression(ufl.Jacobian(mesh), q_points)
        detJ = dolfinx.fem.Expression(ufl.JacobianDeterminant(mesh), q_points)
        K = dolfinx.fem.Expression(ufl.JacobianInverse(mesh), q_points)
        jacs = jacobian.eval(mesh, np.arange(num_cells, dtype=np.int32)).reshape(
            num_cells * num_points, mesh.geometry.dim, mesh.topology.dim
        )
        detJs = detJ.eval(mesh, np.arange(num_cells, dtype=np.int32)).flatten()
        Ks = K.eval(mesh, np.arange(num_cells, dtype=np.int32)).reshape(
            num_cells * num_points, mesh.geometry.dim, mesh.topology.dim
        )

        for i in range(V.dofmap.bs * V.dofmap.dof_layout.num_dofs):
            for q in range(Q.dofmap.bs):
                new_array[:, q * Q_vs : (q + 1) * Q_vs, i] = pull_back(
                    array_evaluated[:, :, q * Q_vs : (q + 1) * Q_vs, i], jacs, detJs, Ks
                ).reshape(num_cells * num_points, Q_vs)
        new_array = new_array.reshape(
            num_cells, num_points, Q.dofmap.bs * Q_vs, V.dofmap.bs * V.dofmap.dof_layout.num_dofs
        )
    else:
        new_array = array_evaluated.reshape(
            num_cells, num_points, Q.dofmap.bs * Q_vs, V.dofmap.bs * V.dofmap.dof_layout.num_dofs
        )

    interpolated_matrix = np.zeros(
        (
            num_cells,
            Q.dofmap.dof_layout.num_dofs * Q.dofmap.bs,
            V.dofmap.bs * V.dofmap.dof_layout.num_dofs,
        ),
        dtype=np.float64,
    )
    # Check if interpolation matrix of dual operator is identity, then we can use a vectorized
    # version of this step
    if Q.element.interpolation_ident:
        # Smart vectorized version with identity mapping
        if Q.dofmap.bs == 1:
            interpolated_matrix = new_array.transpose(0, 2, 1, 3).reshape(
                new_array.shape[0], new_array.shape[1] * new_array.shape[2], new_array.shape[3]
            )
        else:
            i_scalar = new_array.transpose(0, 2, 1, 3)
            interpolated_matrix = np.zeros(
                (new_array.shape[0], new_array.shape[1] * new_array.shape[2], new_array.shape[3])
            )
            for q in range(Q.dofmap.bs):
                interpolated_matrix[:, q :: Q.dofmap.bs, :] = i_scalar[:, q, :, :]

    else:
        # Tedious non-identity version
        for c in range(num_cells):
            for i in range(V.dofmap.bs * V.dofmap.dof_layout.num_dofs):
                tmp_array = np.zeros((int(num_points), Q.dofmap.bs * Q_vs), dtype=np.float64)
                for p in range(num_points):
                    tmp_array[p] = new_array[c, p, :, i]
                if Q.dofmap.bs == 1:
                    interpolated_matrix[c, :, i] = (im @ tmp_array.T.flatten()).flatten()
                else:
                    for q in range(Q.dofmap.bs):
                        interpolated_matrix[c, q :: Q.dofmap.bs, i] = (
                            im @ tmp_array.T[q].flatten()
                        ).flatten()
    # Apply dof transformation to each column (using Piopla maps)
    mesh.topology.create_entity_permutations()
    if Q.element.needs_dof_transformations:
        cell_perm = mesh.topology.get_cell_permutation_info()[:num_cells]

        permuted_matrix = interpolated_matrix.flatten().copy()
        Q.element.Tt_inv_apply(
            permuted_matrix, cell_perm, V.dofmap.bs * V.dofmap.dof_layout.num_dofs
        )
    else:
        permuted_matrix = interpolated_matrix.flatten()
    return permuted_matrix.reshape(interpolated_matrix.shape)


def interpolation_matrix(
    expr: ufl.core.expr.Expr, Q: dolfinx.fem.FunctionSpace
) -> dolfinx.la.MatrixCSR:
    """Create the interpolation matrix :math:`\\Lambda` of a
    :py:class:`UFL-expression<ufl.core.expr.Expr>` such that

    .. math::
        \\begin{align*}
        \\Lambda: V &\\rightarrow Q \\\\
        \\Lambda u &= \\sum_{i=0}^{N_Q-1}\\sum_{j=0}^{N_V-1} \\phi_i l_i(expr(\\psi_j))u_j
        \\end{align*}

    where :math:`l_j` is the dual basis of the space :math:`Q` with
    basis functions :math:`\\phi_j`, and :math:`\\psi_j` are the basis functions of the
    space :math:`V`.

    Args:
        expr: The UFL expression
        Q: Output interpolation space

    Returns:
        Interpolation matrix as a :py:class:`MatrixCSR<dolfinx.la.MatrixCSR>`.
    """

    arguments = ufl.algorithms.extract_arguments(expr)
    assert len(arguments) == 1
    V = arguments[0].ufl_function_space()

    interpolation_data = prepare_interpolation_data(expr, Q)

    q = ufl.TestFunction(Q)
    a = dolfinx.fem.form(ufl.inner(expr, q) * ufl.dx)

    def scatter(
        A: dolfinx.la.MatrixCSR,
        num_cells: int,
        dofs_visited: npt.NDArray[np.int32],
        num_rows_local: int,
        array_evaluated: npt.NDArray[np.inexact],
        dofmap0: npt.NDArray[np.int32],
        dofmap1: npt.NDArray[np.int32],
    ):
        A.data[:] = 0
        for i in range(num_cells):
            rows = dofmap0[i, :]
            cols = dofmap1[i, :]
            A_local = array_evaluated[i].reshape(len(rows), len(cols))
            row_filter = (dofs_visited[rows] == 1) | (rows >= num_rows_local)
            A_local[row_filter] = 0
            A.add(A_local.flatten(), rows, cols)
            dofs_visited[rows] = 1

    A = dolfinx.fem.create_matrix(a)  # , dolfinx.la.BlockMode.expanded)

    row_dofmap = unroll_dofmap(Q.dofmap.list, Q.dofmap.bs)  # (num_cells, num_rows)
    col_dofmap = unroll_dofmap(V.dofmap.list, V.dofmap.bs)  # (num_cells, num_cols)

    num_cells = Q.mesh.topology.index_map(Q.mesh.topology.dim).size_local
    dofs_visited = np.zeros(
        (Q.dofmap.index_map.size_local + Q.dofmap.index_map.num_ghosts) * Q.dofmap.index_map_bs,
        dtype=np.int8,
    )
    num_rows_local = Q.dofmap.index_map.size_local * Q.dofmap.bs
    scatter(A, num_cells, dofs_visited, num_rows_local, interpolation_data, row_dofmap, col_dofmap)
    A.scatter_reverse()
    return A


if dolfinx.has_petsc4py:

    def petsc_interpolation_matrix(
        expr: ufl.core.expr.Expr, Q: dolfinx.fem.FunctionSpace, use_petsc: bool = False
    ) -> PETSc.Mat:
        """Create the interpolation matrix :math:`\\Lambda` of a
        :py:class:`UFL-expression<ufl.core.expr.Expr>` such that

        .. math::
            \\begin{align*}
            \\Lambda: V &\\rightarrow Q \\\\
            \\Lambda u &= \\sum_{i=0}^{N_Q-1}\\sum_{j=0}^{N_V-1} \\phi_i l_i(expr(\\psi_j))u_j
            \\end{align*}

        where :math:`l_j` is the dual basis of the space :math:`Q` with basis
        functions :math:`\\phi_j`, and :math:`\\psi_j` are the basis functions
        of the space :math:`V`.

        Args:
            expr: The UFL expression
            Q: Output interpolation space

        Returns:
            Interpolation matrix as a :py:class:`PETSc.Mat<petsc4py.PETSc.Mat>`.
        """
        arguments = ufl.algorithms.extract_arguments(expr)
        assert len(arguments) == 1
        V = arguments[0].ufl_function_space()

        interpolation_data = prepare_interpolation_data(expr, Q)

        q = ufl.TestFunction(Q)
        a = dolfinx.fem.form(ufl.inner(expr, q) * ufl.dx)
        A = dolfinx.fem.petsc.create_matrix(a)

        def scatter(
            A: PETSc.Mat,
            num_cells: int,
            dofs_visited: npt.NDArray[np.int32],
            num_rows_local: int,
            array_evaluated: npt.NDArray[np.inexact],
            dofmap0: npt.NDArray[np.int32],
            dofmap1: npt.NDArray[np.int32],
        ):
            A.zeroEntries()
            for i in range(num_cells):
                rows = dofmap0[i, :]
                cols = dofmap1[i, :]
                A_local = array_evaluated[i].reshape(len(rows), len(cols))
                row_filter = (dofs_visited[rows] == 1) | (rows >= num_rows_local)
                A_local[row_filter] = 0
                A.setValuesLocal(rows, cols, A_local, addv=PETSc.InsertMode.ADD_VALUES)
                dofs_visited[rows] = 1

        row_dofmap = unroll_dofmap(Q.dofmap.list, Q.dofmap.bs)  # (num_cells, num_rows)
        col_dofmap = unroll_dofmap(V.dofmap.list, V.dofmap.bs)  # (num_cells, num_cols)
        num_cells = Q.mesh.topology.index_map(Q.mesh.topology.dim).size_local
        dofs_visited = np.zeros(
            (Q.dofmap.index_map.size_local + Q.dofmap.index_map.num_ghosts) * Q.dofmap.index_map_bs,
            dtype=np.int8,
        )
        num_rows_local = Q.dofmap.index_map.size_local * Q.dofmap.bs
        scatter(
            A, num_cells, dofs_visited, num_rows_local, interpolation_data, row_dofmap, col_dofmap
        )
        A.assemble()
        return A


def interpolate_to_surface_submesh(
    u_volume: dolfinx.fem.Function,
    u_surface: dolfinx.fem.Function,
    submesh_facets: npt.NDArray[np.int32],
    integration_entities: npt.NDArray[np.int32],
):
    """
    Interpolate a function `u_volume` into the function `u_surface`.
    Note:
        Does not work for DG as no dofs are associated with the facets
    Args:
        u_volume: Function to interpolate data from
        u_surface: Function to interpolate data to
        submesh_facets: Cells in facet mesh
        integration_entities: Integration entities on the parent mesh
            corresponding to the facets in `submesh_facets`
    """
    if Version(dolfinx.__version__) < Version("0.10.0"):
        raise RuntimeError("interpolate_to_submesh requires dolfinx version 0.10.0 or higher")
    V_vol = u_volume.function_space
    mesh = V_vol.mesh
    V_surf = u_surface.function_space
    submesh = V_surf.mesh
    ip = V_surf.element.interpolation_points

    expr = dolfinx.fem.Expression(u_volume, ip)

    mesh.topology.create_connectivity(mesh.topology.dim, submesh.topology.dim)
    mesh.topology.create_connectivity(submesh.topology.dim, mesh.topology.dim)

    data = expr.eval(mesh, integration_entities)

    submesh.topology.create_entity_permutations()
    mesh.topology.create_entity_permutations()
    cell_info = mesh.topology.get_cell_permutation_info()
    ft = V_surf.element.basix_element.cell_type

    for i in range(integration_entities.shape[0]):
        perm = np.arange(data.shape[1], dtype=np.int32)
        V_vol.element.basix_element.permute_subentity_closure_inv(
            perm,
            cell_info[integration_entities[i, 0]],
            ft,
            int(integration_entities[i, 1]),
        )
        data[i] = data[i][perm]

    if len(data.shape) == 3:
        # Data is now (num_cells, value_size,num_points)
        data = data.swapaxes(1, 2)
        # Data is now (value_size, num_cells, num_points)
        data = data.swapaxes(0, 1)

    if expr.value_size == 1:
        shaped_data = data.flatten()
    else:
        shaped_data = data.reshape(expr.value_size, -1)

    if hasattr(u_surface._cpp_object, "interpolate_f"):
        interpolate_func = u_surface._cpp_object.interpolate_f
    else:
        interpolate_func = u_surface._cpp_object.interpolate
    interpolate_func(shaped_data, submesh_facets)
    u_surface.x.scatter_forward()



from mpi4py import MPI
import pytest
import dolfinx
from scifem import vertex_to_dofmap, dof_to_vertexmap
import numpy as np


@pytest.mark.parametrize("degree", range(1, 4))
@pytest.mark.parametrize("value_size", [(), (2,), (2, 3)])
def test_vertex_to_dofmap_P(degree, value_size):
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 2)
    V = dolfinx.fem.functionspace(mesh, ("Lagrange", degree, value_size))

    v_to_d = vertex_to_dofmap(V)
    mesh.topology.create_connectivity(0, mesh.topology.dim)
    geom_indices = dolfinx.mesh.entities_to_geometry(
        mesh, 0, np.arange(len(v_to_d), dtype=np.int32)
    )

    x_V = V.tabulate_dof_coordinates()
    x_g = mesh.geometry.x
    tol = 1e3 * np.finfo(x_g.dtype).eps
    mesh = V.mesh

    np.testing.assert_allclose(x_V[v_to_d], x_g[geom_indices.flatten()], atol=tol)

    # Test inverse map
    d_to_v = dof_to_vertexmap(V)
    np.testing.assert_allclose(d_to_v[v_to_d], np.arange(len(v_to_d)))

    # Mark all dofs connected with a vertex
    dof_marker = np.full(len(d_to_v), True, dtype=np.bool_)
    dof_marker[v_to_d] = False

    other_dofs = np.flatnonzero(dof_marker)
    np.testing.assert_allclose(d_to_v[other_dofs], -1)



# # Entity markers
#
# Author: Jørgen S. Dokken
#
# SPDX-License-Identifier: MIT
#
# DOLFINx gives you full control for marking entities.
# However, sometimes this can feel a bit repetative.
# In this example we will show how to use {py:func}`scifem.create_entity_markers`.

from mpi4py import MPI
import dolfinx
import numpy as np

# We start by creating a simple unit square

mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 60, 60)

# Next, we want to mark some of the cells in our domain.
# We create three marker functions below.
# Each of them takes in a ``(3, num_points)`` array, and returns a boolean array of size ``num_points``.


def left(x):
    return x[0] < 0.2


def right(x):
    return x[0] > 0.9


def inner(x):
    # We use numpy bit operator `&` for "and"
    return (x[0] > 0.3) & (x[0] < 0.7)


# We want to mark these entities with  with unique integers 1, 3 and 7.

from scifem import create_entity_markers

cell_tag = create_entity_markers(mesh, mesh.topology.dim, [(1, left), (3, right), (7, inner)])

# Next we can plot these marked entities

import pyvista

vtk_grid = dolfinx.plot.vtk_mesh(mesh, cell_tag.dim, cell_tag.indices)
grid = pyvista.UnstructuredGrid(*vtk_grid)
grid.cell_data["Marker"] = cell_tag.values

# Create plotter

plotter = pyvista.Plotter()
plotter.add_mesh(grid)
plotter.view_xy()
if not pyvista.OFF_SCREEN:
    plotter.show()


# We can also mark lower order entities, such as facets


def circle(x):
    return x[0] ** 2 + x[1] ** 2 <= 0.16**2


def top(x):
    return x[1] > 0.9


facet_tags = create_entity_markers(mesh, mesh.topology.dim - 1, [(2, top), (7, circle)])


facet_grid = dolfinx.plot.vtk_mesh(mesh, facet_tags.dim, facet_tags.indices)

fgrid = pyvista.UnstructuredGrid(*facet_grid)
fgrid.cell_data["Marker"] = facet_tags.values

fplotter = pyvista.Plotter()
fplotter.add_mesh(fgrid)
fplotter.view_xy()
if not pyvista.OFF_SCREEN:
    fplotter.show()

# We can also exclude interior facets by adding `on_boundary: True` (by default this is set to False).

boundary_facet_tags = create_entity_markers(
    mesh, mesh.topology.dim - 1, [(2, top, True), (7, circle, False)]
)


boundary_grid = dolfinx.plot.vtk_mesh(mesh, boundary_facet_tags.dim, boundary_facet_tags.indices)

bfgrid = pyvista.UnstructuredGrid(*boundary_grid)
bfgrid.cell_data["Marker"] = boundary_facet_tags.values

bfplotter = pyvista.Plotter()
bfplotter.add_mesh(bfgrid)
bfplotter.view_xy()
if not pyvista.OFF_SCREEN:
    bfplotter.show()



from mpi4py import MPI
import dolfinx
import dolfinx.fem.petsc
import basix
import ufl
import numpy as np
import scifem
import pytest
from packaging.version import Version
from petsc4py import PETSc
from unittest.mock import patch

try:
    import h5py  # noqa: F401

    hash5py = True
except ImportError:
    hash5py = False


if h5py.h5.HDF5_VERSION_COMPILED_AGAINST == (2, 0, 0):
    pytest.skip(
        "h5py and adios2 built against HDF5 2.0.0 does not work properly", allow_module_level=True
    )


stype = PETSc.ScalarType
rtype = PETSc.RealType
_dolfinx_version = Version(dolfinx.__version__)


@pytest.mark.parametrize("use_ctx_manager", [True, False])
@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("h5py", marks=pytest.mark.skipif(not hash5py, reason="h5py is not installed")),
        "adios2",
    ],
)
@pytest.mark.parametrize("degree", [1, 2, 3])
@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
@pytest.mark.parametrize("value_shape", [(), (2,)])
def test_XDMFFile_2D_dolfinx(cell_type, degree, value_shape, backend, use_ctx_manager, tmp_path):
    folder = MPI.COMM_WORLD.bcast(tmp_path, root=0)

    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10, cell_type, dtype=np.float64)

    el = basix.ufl.quadrature_element(
        scheme="default", degree=degree, cell=mesh.basix_cell(), value_shape=value_shape
    )
    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V, dtype=stype)
    v = dolfinx.fem.Function(V, dtype=stype)
    t = dolfinx.fem.Constant(mesh, stype(0.0))

    X = ufl.SpatialCoordinate(mesh)
    if value_shape == ():
        u_expr = ufl.sin(ufl.pi * X[0]) * ufl.sin(ufl.pi * X[1]) * ufl.cos(ufl.pi * t)
        v_expr = ufl.cos(ufl.pi * X[0]) * ufl.cos(ufl.pi * X[1]) * ufl.cos(ufl.pi * t)

    else:
        u_expr = X * ufl.cos(ufl.pi * t)
        v_expr = X * ufl.sin(X[0]) * ufl.cos(ufl.pi * t)

    u.name = "u"
    v.name = "v"

    if _dolfinx_version >= Version("0.10"):
        points = V.element.interpolation_points
    else:
        points = V.element.interpolation_points()

    if use_ctx_manager:
        with scifem.xdmf.XDMFFile(folder / "data.xdmf", [u, v], backend=backend) as xdmf:
            u.interpolate(dolfinx.fem.Expression(u_expr, points))
            v.interpolate(dolfinx.fem.Expression(v_expr, points))
            xdmf.write(rtype(t.value))

            t.value = 0.3
            u.interpolate(dolfinx.fem.Expression(u_expr, points))
            v.interpolate(dolfinx.fem.Expression(v_expr, points))
            xdmf.write(rtype(t.value))
    else:
        xdmf = scifem.xdmf.XDMFFile(folder / "data.xdmf", [u, v], backend=backend)
        u.interpolate(dolfinx.fem.Expression(u_expr, points))
        v.interpolate(dolfinx.fem.Expression(v_expr, points))
        xdmf.write(rtype(t.value))

        t.value = 0.3
        u.interpolate(dolfinx.fem.Expression(u_expr, points))
        v.interpolate(dolfinx.fem.Expression(v_expr, points))
        xdmf.write(rtype(t.value))

        xdmf.close()

    assert (folder / "data.xdmf").is_file()
    assert (folder / "data.h5").is_file()


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("h5py", marks=pytest.mark.skipif(not hash5py, reason="h5py is not installed")),
        "adios2",
    ],
)
@pytest.mark.parametrize("degree", [1, 2, 3])
@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
@pytest.mark.parametrize("value_shape", [(), (2,)])
def test_XDMFFile_2D_numpy(cell_type, degree, value_shape, backend, tmp_path):
    folder = MPI.COMM_WORLD.bcast(tmp_path, root=0)

    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10, cell_type, dtype=np.float64)

    el = basix.ufl.quadrature_element(
        scheme="default", degree=degree, cell=mesh.basix_cell(), value_shape=value_shape
    )
    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V, dtype=stype)
    v = dolfinx.fem.Function(V, dtype=stype)
    t = dolfinx.fem.Constant(mesh, stype(0.0))

    X = ufl.SpatialCoordinate(mesh)
    if value_shape == ():
        u_expr = ufl.sin(ufl.pi * X[0]) * ufl.sin(ufl.pi * X[1]) * ufl.cos(ufl.pi * t)
        v_expr = ufl.cos(ufl.pi * X[0]) * ufl.cos(ufl.pi * X[1]) * ufl.cos(ufl.pi * t)

    else:
        u_expr = X * ufl.cos(ufl.pi * t)
        v_expr = X * ufl.sin(X[0]) * ufl.cos(ufl.pi * t)

    u.name = "u"
    v.name = "v"

    if _dolfinx_version >= Version("0.10"):
        points = V.element.interpolation_points
    else:
        points = V.element.interpolation_points()

    function_space_data = scifem.xdmf.create_function_space_data(u.function_space)

    with scifem.xdmf.NumpyXDMFFile(
        folder / "data.xdmf",
        [u.x.array, v.x.array],
        function_space_data=function_space_data,
        backend=backend,
    ) as xdmf:
        u.interpolate(dolfinx.fem.Expression(u_expr, points))
        v.interpolate(dolfinx.fem.Expression(v_expr, points))
        xdmf.write(rtype(t.value))

        t.value = 0.3
        u.interpolate(dolfinx.fem.Expression(u_expr, points))
        v.interpolate(dolfinx.fem.Expression(v_expr, points))
        xdmf.write(rtype(t.value))

    assert (folder / "data.xdmf").is_file()
    assert (folder / "data.h5").is_file()


@pytest.mark.parametrize("use_ctx_manager", [True, False])
@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("h5py", marks=pytest.mark.skipif(not hash5py, reason="h5py is not installed")),
        "adios2",
    ],
)
@pytest.mark.parametrize("degree", [1, 2, 3])
@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.tetrahedron, dolfinx.mesh.CellType.hexahedron]
)
@pytest.mark.parametrize("value_shape", [(), (3,)])
def test_XDMFFile_3D(cell_type, degree, value_shape, backend, use_ctx_manager, tmp_path):
    folder = MPI.COMM_WORLD.bcast(tmp_path, root=0)

    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 2, 3, 5, cell_type, dtype=np.float64)

    el = basix.ufl.quadrature_element(
        scheme="default", degree=degree, cell=mesh.basix_cell(), value_shape=value_shape
    )
    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V, dtype=stype)
    v = dolfinx.fem.Function(V, dtype=stype)
    t = dolfinx.fem.Constant(mesh, stype(0.0))

    X = ufl.SpatialCoordinate(mesh)
    if value_shape == ():
        u_expr = ufl.sin(ufl.pi * X[0]) * ufl.sin(ufl.pi * X[1]) * ufl.cos(ufl.pi * t)
        v_expr = ufl.cos(ufl.pi * X[0]) * ufl.cos(ufl.pi * X[1]) * ufl.cos(ufl.pi * t)

    else:
        u_expr = X * ufl.cos(ufl.pi * t)
        v_expr = X * ufl.sin(X[0]) * ufl.cos(ufl.pi * t)

    u.name = "u"
    v.name = "v"

    if _dolfinx_version >= Version("0.10"):
        points = V.element.interpolation_points
    else:
        points = V.element.interpolation_points()

    if use_ctx_manager:
        with scifem.xdmf.XDMFFile(folder / "data.xdmf", [u, v], backend=backend) as xdmf:
            u.interpolate(dolfinx.fem.Expression(u_expr, points))
            v.interpolate(dolfinx.fem.Expression(v_expr, points))
            xdmf.write(rtype(t.value))

            t.value = 0.3
            u.interpolate(dolfinx.fem.Expression(u_expr, points))
            v.interpolate(dolfinx.fem.Expression(v_expr, points))
            xdmf.write(rtype(t.value))
    else:
        xdmf = scifem.xdmf.XDMFFile(folder / "data.xdmf", [u, v], backend=backend)
        u.interpolate(dolfinx.fem.Expression(u_expr, points))
        v.interpolate(dolfinx.fem.Expression(v_expr, points))
        xdmf.write(rtype(t.value))

        t.value = 0.3
        u.interpolate(dolfinx.fem.Expression(u_expr, points))
        v.interpolate(dolfinx.fem.Expression(v_expr, points))
        xdmf.write(rtype(t.value))

        xdmf.close()

    assert (folder / "data.xdmf").is_file()
    assert (folder / "data.h5").is_file()


@pytest.mark.parametrize("degree", [1, 2, 3])
@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
@pytest.mark.parametrize("value_shape", [(), (2,)])
def test_create_ponitcloud_2D(cell_type, degree, value_shape, tmp_path):
    folder = MPI.COMM_WORLD.bcast(tmp_path, root=0)

    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10, cell_type, dtype=np.float64)

    el = basix.ufl.quadrature_element(
        scheme="default", degree=degree, cell=mesh.basix_cell(), value_shape=value_shape
    )
    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V)
    v = dolfinx.fem.Function(V)

    if value_shape == ():
        u.interpolate(lambda x: np.sin(np.pi * x[0]) * np.sin(np.pi * x[1]))
        v.interpolate(lambda x: np.cos(np.pi * x[0]) * np.cos(np.pi * x[1]))
    else:
        u.interpolate(lambda x: x[:2, :])
        v.interpolate(lambda x: 2 * x[:2, :])
    u.name = "u"

    v.name = "v"

    scifem.xdmf.create_pointcloud(folder / "data.xdmf", [u, v])
    assert (folder / "data.xdmf").is_file()
    assert (folder / "data.h5").is_file()


@pytest.mark.parametrize("degree", [1, 2, 3])
@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.tetrahedron, dolfinx.mesh.CellType.hexahedron]
)
@pytest.mark.parametrize("value_shape", [(), (3,)])
def test_create_pointcloud_3D(cell_type, degree, value_shape, tmp_path):
    folder = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 2, 3, 5, cell_type, dtype=np.float64)

    el = basix.ufl.quadrature_element(
        scheme="default", degree=degree, cell=mesh.basix_cell(), value_shape=value_shape
    )
    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V)
    v = dolfinx.fem.Function(V)

    if value_shape == ():
        u.interpolate(lambda x: np.sin(np.pi * x[0]) * np.sin(np.pi * x[1]) * np.sin(np.pi * x[2]))
        v.interpolate(lambda x: np.cos(np.pi * x[0]) * np.cos(np.pi * x[1]) * np.cos(np.pi * x[2]))
    else:
        u.interpolate(lambda x: x[:3, :])
        v.interpolate(lambda x: 2 * x[:3, :])

    u.name = "u"
    v.name = "v"

    scifem.xdmf.create_pointcloud(folder / "data.xdmf", [u, v])
    assert (folder / "data.xdmf").is_file()
    assert (folder / "data.h5").is_file()


def test_different_function_spaces_raises_ValueError(tmp_path):
    folder = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    mesh = dolfinx.mesh.create_unit_cube(
        MPI.COMM_WORLD, 2, 3, 5, dolfinx.mesh.CellType.tetrahedron, dtype=np.float64
    )
    V1 = dolfinx.fem.functionspace(mesh, ("Lagrange", 1))
    V2 = dolfinx.fem.functionspace(mesh, ("Lagrange", 2))

    with pytest.raises(ValueError):
        scifem.xdmf.create_pointcloud(
            folder / "data.xdmf", [dolfinx.fem.Function(V1), dolfinx.fem.Function(V2)]
        )


@pytest.mark.parametrize("degree", [1, 2, 3])
@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.tetrahedron, dolfinx.mesh.CellType.hexahedron]
)
@pytest.mark.parametrize("value_shape", [(), (3,)])
@pytest.mark.skipif(not hash5py, reason="h5py not installed")
def test_h5py_fallback_3D(cell_type, degree, value_shape, tmp_path):
    folder = MPI.COMM_WORLD.bcast(tmp_path, root=0)

    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 2, 3, 5, cell_type, dtype=np.float64)

    el = basix.ufl.quadrature_element(
        scheme="default", degree=degree, cell=mesh.basix_cell(), value_shape=value_shape
    )
    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V)
    v = dolfinx.fem.Function(V)

    if value_shape == ():
        u.interpolate(lambda x: np.sin(np.pi * x[0]) * np.sin(np.pi * x[1]) * np.sin(np.pi * x[2]))
        v.interpolate(lambda x: np.cos(np.pi * x[0]) * np.cos(np.pi * x[1]) * np.cos(np.pi * x[2]))
    else:
        u.interpolate(lambda x: x[:3, :])
        v.interpolate(lambda x: 2 * x[:3, :])

    u.name = "u"
    v.name = "v"

    with patch.dict("sys.modules", {"adios2": None}):
        scifem.xdmf.create_pointcloud(folder / "data.xdmf", [u, v])
    assert (folder / "data.xdmf").is_file()
    assert (folder / "data.h5").is_file()



# # Nonlinear elasticity with blocked Newton solver
#
# Author: Henrik N. T. Finsberg
#
# SPDX-License-Identifier: MIT

# In this example we will solve a nonlinear elasticity problem using a blocked Newton solver.
# We consider a unit cube domain $\Omega = [0, 1]^3$ with Dirichlet boundary conditions on the left face and traction force on the right face, and we seek a displacement field $\mathbf{u}: \Omega \to \mathbb{R}^3$ that solves the momentum balance equation
#
# $$
# \begin{align}
# \nabla \cdot \mathbf{P} = 0, \quad \mathbf{X} \in \Omega, \\
# \end{align}
# $$
#
# where $\mathbf{P}$ is the first Piola-Kirchhoff stress tensor, and $\mathbf{X}$ is the reference configuration. We consider a Neo-Hookean material model, where the strain energy density is given by
#
# $$
# \begin{align}
# \psi = \frac{\mu}{2}(\text{tr}(\mathbf{C}) - 3),
# \end{align}
# $$
#
# and the first Piola-Kirchhoff stress tensor is given by
#
# $$
# \begin{align}
# \mathbf{P} = \frac{\partial \psi}{\partial \mathbf{F}}
# \end{align}
# $$
#
# where $\mathbf{F} = \nabla \mathbf{u} + \mathbf{I}$ is the deformation gradient, $\mathbf{C} = \mathbf{F}^T \mathbf{F}$ is the right Cauchy-Green tensor, $\mu$ is the shear modulus, and $p$ is the pressure.
# We also enforce the incompressibility constraint
#
# $$
# \begin{align}
# J = \det(\mathbf{F}) = 1,
# \end{align}
# $$
#
# so that the total Lagrangian is given by
#
# $$
# \begin{align}
# \mathcal{L}(\mathbf{u}, p) = \int_{\Omega} \psi \, dx - \int_{\partial \Omega} t \cdot \mathbf{u} \, ds + \int_{\Omega} p(J - 1) \, dx.
# \end{align}
# $$
#
# Here $t$ is the traction force which is set to $10$ on the right face of the cube and $0$ elsewhere.
# The Euler-Lagrange equations for this problem are given by: Find $\mathbf{u} \in V$ and $p \in Q$ such that
#
# $$
# \begin{align}
# D_{\delta \mathbf{u} } \mathcal{L}(\mathbf{u}, p) = 0, \quad \forall \delta \mathbf{u} \in V, \\
# D_{\delta p} \mathcal{L}(\mathbf{u}, p) = 0, \quad \forall \delta p \in Q,
# \end{align}
# $$
#
# where $V$ is the displacement space and $Q$ is the pressure space. For this we select $Q_2/P_1$ elements i.e second order Lagrange elements for $\mathbf{u}$ and [first order discontinuous polynomial cubical elements](https://defelement.org/elements/examples/quadrilateral-dpc-1.html) for $p$, which is a stable element for incompressible elasticity {cite}`auricchio2013approximation`.
# Note also that the Euler-Lagrange equations can be derived automatically using `ufl`.
#
#

import logging
from mpi4py import MPI
import numpy as np
import ufl
import dolfinx
import scifem

# Initialize logging and set log level to info

logging.basicConfig(level=logging.INFO)

# We create the mesh and the function spaces

mesh = dolfinx.mesh.create_unit_cube(
    MPI.COMM_WORLD, 3, 3, 3, dolfinx.mesh.CellType.hexahedron, dtype=np.float64
)
V = dolfinx.fem.functionspace(mesh, ("Lagrange", 2, (3,)))
Q = dolfinx.fem.functionspace(mesh, ("DPC", 1,))

# And the test and trial functions
#

v = ufl.TestFunction(V)
q = ufl.TestFunction(Q)
du = ufl.TrialFunction(V)
dp = ufl.TrialFunction(Q)
u = dolfinx.fem.Function(V)
p = dolfinx.fem.Function(Q)


# Next we create the facet tags for the left and right faces

def left(x):
    return np.isclose(x[0], 0)


def right(x):
    return np.isclose(x[0], 1)


facet_tags = scifem.create_entity_markers(
    mesh, mesh.topology.dim - 1, [(1, left, True), (2, right, True)]
)

# We create the Dirichlet boundary conditions on the left face

facets_left = facet_tags.find(1)
dofs_left = dolfinx.fem.locate_dofs_topological(V, 2, facets_left)
u_bc_left = dolfinx.fem.Function(V)
u_bc_left.x.array[:] = 0
bc = dolfinx.fem.dirichletbc(u_bc_left, dofs_left)

# Define the deformation gradient, right Cauchy-Green tensor, and invariants of the deformation tensors

d = len(u)
I = ufl.Identity(d)             # Identity tensor
F = I + ufl.grad(u)             # Deformation gradient
C = F.T*F                       # Right Cauchy-Green tensor
I1 = ufl.tr(C)                  # First invariant of C
J  = ufl.det(F)                 # Jacobian of F

# Traction for to be applied on the right face

t = dolfinx.fem.Constant(mesh, dolfinx.default_scalar_type(10.0))

N = ufl.FacetNormal(mesh)

# Material parameters and strain energy density
mu = dolfinx.fem.Constant(mesh, 10.0)
psi = (mu / 2)*(I1 - 3)

# We for the total Lagrangian

L = psi*ufl.dx - ufl.inner(t * N, u)*ufl.ds(subdomain_data=facet_tags, subdomain_id=2)  + p * (J - 1) * ufl.dx

# and take the first variation of the total Lagrangian to obtain the residual

r_u = ufl.derivative(L, u, v)
r_p = ufl.derivative(L, p, q)
R = [r_u, r_p]

# We do the same for the second variation to obtain the Jacobian

K = [
    [ufl.derivative(r_u, u, du), ufl.derivative(r_u, p, dp)],
    [ufl.derivative(r_p, u, du), ufl.derivative(r_p, p, dp)],
]


# Now we can create the Newton solver and solve the problem

petsc_options = {"ksp_type": "preonly", "pc_type": "lu", "pc_factor_mat_solver_type": "mumps"}
solver = scifem.NewtonSolver(R, K, [u, p], max_iterations=25, bcs=[bc], petsc_options=petsc_options)

# We can also set a callback function that is called before and after the solve, which takes the solver object as input


def pre_solve(solver: scifem.NewtonSolver):
    print(f"Starting solve with {solver.max_iterations} iterations")


def post_solve(solver: scifem.NewtonSolver):
    print(f"Solve completed in with correction norm {solver.dx.norm(0)}")


solver.set_pre_solve_callback(pre_solve)
solver.set_post_solve_callback(post_solve)

solver.solve()

# Finally, we can visualize the solution using `pyvista`

import pyvista
p = pyvista.Plotter()
topology, cell_types, geometry = dolfinx.plot.vtk_mesh(V)
grid = pyvista.UnstructuredGrid(topology, cell_types, geometry)
linear_grid = pyvista.UnstructuredGrid(*dolfinx.plot.vtk_mesh(mesh))
grid["u"] = u.x.array.reshape((geometry.shape[0], 3))
p.add_mesh(linear_grid, style="wireframe", color="k")
warped = grid.warp_by_vector("u", factor=1.5)
p.add_mesh(warped, show_edges=False)
p.show_axes()
if not pyvista.OFF_SCREEN:
    p.show()
else:
    figure_as_array = p.screenshot("displacement.png")


# # References
# ```{bibliography}
# ```
#



from mpi4py import MPI
import dolfinx
import numpy as np
from scifem.biomedical import read_mri_data_to_function, read_mri_data_to_tag
import pytest
import basix.ufl


@pytest.mark.parametrize("degree", [0, 1, 3])
@pytest.mark.parametrize("M", [25, 17])
@pytest.mark.parametrize("Nx", [7, 3])
@pytest.mark.parametrize("Ny", [23, 8])
@pytest.mark.parametrize("Nz", [9, 17])
@pytest.mark.parametrize("theta", [np.pi / 3, 0, -np.pi])
@pytest.mark.parametrize("translation", [np.array([0, 0, 0]), np.array([2.1, 1.3, 0.4])])
@pytest.mark.parametrize("mri_data_format", ["nifti", "mgh"])
@pytest.mark.parametrize("external_affine", [True, False])
@pytest.mark.parametrize("use_tkr", [False])
def test_read_mri_data_to_function(
    degree, M, Nx, Ny, Nz, theta, translation, tmp_path, mri_data_format, external_affine, use_tkr
):
    nibabel = pytest.importorskip("nibabel")

    # Generate rotation and scaling matrix equivalent to a unit cube
    rotation_matrix_3D = np.array(
        [[1, 0, 0], [0, np.cos(theta), -np.sin(theta)], [0, np.sin(theta), np.cos(theta)]]
    )
    scale_matrix = 1.0 / M * np.identity(3)

    # Generate the affine mapping for nibabel
    A = np.append(np.dot(rotation_matrix_3D, scale_matrix), (translation).reshape(3, 1), axis=1)
    A = np.vstack([A, [0, 0, 0, 1]])
    Id = np.identity(4)

    # Write transformed data to file
    path = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    if mri_data_format == "nifti":
        data = np.arange(1, M**3 + 1, dtype=np.float64).reshape(M, M, M)
        image = nibabel.nifti1.Nifti1Image(data, affine=Id if external_affine else A)
    elif mri_data_format == "mgh":
        data = np.arange(1, M**3 + 1, dtype=np.int32).reshape(M, M, M)
        image = nibabel.freesurfer.mghformat.MGHImage(data, affine=Id if external_affine else A)

    # Reorient the image to RAS
    orig_ornt = nibabel.io_orientation(image.affine)
    targ_ornt = nibabel.orientations.axcodes2ornt("RAS")
    transform = nibabel.orientations.ornt_transform(orig_ornt, targ_ornt)
    image = image.as_reoriented(transform)
    filename = path.with_suffix(".mgz")

    # Save the image to file
    if MPI.COMM_WORLD.rank == 0:
        nibabel.save(image, filename)
    MPI.COMM_WORLD.Barrier()
    # Create unit cube
    mesh = dolfinx.mesh.create_unit_cube(
        MPI.COMM_WORLD,
        Nx,
        Ny,
        Nz,
        cell_type=dolfinx.cpp.mesh.CellType.hexahedron,
        ghost_mode=dolfinx.mesh.GhostMode.shared_facet,
    )

    # Find voxel position on reference cube (done prior to mesh transformation)
    q_el = basix.ufl.quadrature_element(mesh.topology.cell_name(), degree=degree)
    V = dolfinx.fem.functionspace(mesh, q_el)
    dx = 1 / M
    coords = np.abs(V.tabulate_dof_coordinates())  # Remove rounding error
    # NOTE: Voxel coordinates are based on center of cell rather than corner
    reference_input_index = ((coords) // dx).astype(np.int32)
    reference_data = data[
        reference_input_index[:, 0], reference_input_index[:, 1], reference_input_index[:, 2]
    ]

    # Transform mesh into physical space
    # As the voxel coordinates has origin in the midpoint of the RAS voxel, we add 1/(2M) to
    # the translation to account for this in the mesh, such that for a non-translated geoemtry
    # (theta=0), (translation=0), the origin in physical spaces is at (-0.5,-0.5,-0.5) in RAS,
    # which corresponds to the corner of voxel (0,0,0)
    midpoint_shift = 1.0 / (2 * M) * np.ones(3)
    shifted_coords = mesh.geometry.x - midpoint_shift

    # Rotate, then translate geometry
    mesh.geometry.x[:] = np.dot(rotation_matrix_3D, shifted_coords.T).T + translation

    # # Read data to function
    if external_affine:
        func = read_mri_data_to_function(
            filename, mesh, degree=degree, dtype=np.float64, vox2ras_transform=A, use_tkr=use_tkr
        )
    else:
        func = read_mri_data_to_function(
            filename, mesh, degree=degree, dtype=np.float64, use_tkr=use_tkr
        )

    np.testing.assert_allclose(func.x.array, reference_data)

    if degree == 0:
        cell_map = mesh.topology.index_map(mesh.topology.dim)
        num_cells_local = cell_map.size_local + cell_map.num_ghosts
        # Pick every second cell
        cells = np.arange(num_cells_local, dtype=np.int32)[::2]
        if external_affine:
            tag = read_mri_data_to_tag(
                filename,
                mesh,
                edim=mesh.topology.dim,
                entities=cells,
                vox2ras_transform=A,
                use_tkr=use_tkr,
            )
        else:
            tag = read_mri_data_to_tag(
                filename, mesh, edim=mesh.topology.dim, entities=cells, use_tkr=use_tkr
            )
        np.testing.assert_allclose(tag.values, reference_data[::2])



import dolfinx
import typing

__all__ = [
    "zero_petsc_vector",
    "ghost_update",
    "apply_lifting_and_set_bc",
]

try:
    from petsc4py import PETSc

    def zero_petsc_vector(b):
        """Zero a PETSc vector, including ghosts"""

        if b.getType() == PETSc.Vec.Type.NEST:
            for b_sub in b.getNestSubVecs():
                with b_sub.localForm() as b_local:
                    b_local.set(0.0)
        else:
            with b.localForm() as b_loc:
                b_loc.set(0)

    def ghost_update(x, insert_mode, scatter_mode):
        """Ghost update a vector"""
        if x.getType() == PETSc.Vec.Type.NEST:
            for x_sub in x.getNestSubVecs():
                x_sub.ghostUpdate(addv=insert_mode, mode=scatter_mode)
        else:
            x.ghostUpdate(addv=insert_mode, mode=scatter_mode)

except ImportError:

    def zero_petsc_vector(_b):
        """Zero a PETSc vector, including ghosts"""
        raise RuntimeError("petsc4py is not available. Cannot zero vector.")

    def ghost_update(_x, _insert_mode, _scatter_mode):
        """Ghost update a vector"""
        raise RuntimeError("petsc4py is not available. Cannot ghost update vector.")


if dolfinx.has_petsc4py:
    from petsc4py import PETSc
    import dolfinx.fem.petsc

    def apply_lifting_and_set_bc(
        b: PETSc.Vec,
        a: typing.Union[
            typing.Iterable[dolfinx.fem.Form], typing.Iterable[typing.Iterable[dolfinx.fem.Form]]
        ],
        bcs: typing.Union[
            typing.Iterable[dolfinx.fem.DirichletBC],
            typing.Iterable[typing.Iterable[dolfinx.fem.DirichletBC]],
        ],
        x: typing.Optional[PETSc.Vec] = None,
        alpha: float = 1.0,
    ):
        """Apply lifting to a vector and set boundary conditions.

        Convenience function to apply lifting and set boundary conditions for multiple matrix types.
        This modifies the vector `b` such that

        .. math::
            b =
            \\begin{pmatrix}
              b_{free} \\\\
              b_{bc}
            \\end{pmatrix} =
            \\begin{pmatrix}
              b_{free} - \\alpha \\sum_{i=0}^n a[i] (u_{bc}[i] - x[i])\\\\
              u_{bc}[0]\\\\
              \\vdots\\\\
              u_{bc}[n]
            \\end{pmatrix}.

        where :math:`b_{free}` is the free part of the vector, :math:`b_{bc}` is the part that has
        boundary conditions applied, :math:`u_{bc}[i]` is the value of the ith boundary condition.

        Args:
            b: The vector to apply lifting to.
            a: Sequence fo forms to apply lifting from. If the system is blocked or nested,
                his is a nested list of forms.
            bcs: The boundary conditions to apply. If the form is blocked or nested, this is a list,
                while if it is a single form, this is a nested list.
            x: Vector to subtract from the boundary conditions. Usually used in a Newton iteration.
            alpha: The scaling factor for the boundary conditions.
        """
        if hasattr(dolfinx.fem.petsc, "create_vector_nest"):
            raise NotImplementedError(
                "This function is only implemented for later versions of DOLFINx."
            )

        try:
            bcs0 = dolfinx.fem.bcs_by_block(dolfinx.fem.extract_function_spaces(a, 0), bcs)
            bcs1 = dolfinx.fem.bcs_by_block(dolfinx.fem.extract_function_spaces(a, 1), bcs)
        except AssertionError:
            bcs0 = bcs
            bcs1 = bcs

        dolfinx.fem.petsc.apply_lifting(b, a, bcs=bcs1, x0=x, alpha=alpha)
        ghost_update(b, PETSc.InsertMode.ADD_VALUES, PETSc.ScatterMode.REVERSE)
        try:
            dolfinx.fem.petsc.set_bc(b, bcs0, x0=x, alpha=alpha)
        except AttributeError:
            for _bcs in bcs0:
                dolfinx.fem.petsc.set_bc(b, _bcs, x0=x, alpha=alpha)
        except TypeError:
            # Catch the case when bcs0 shouldn't be a nested list (e.g., for a single form)
            assert len(bcs0) == 1, "bcs0 should be a single DirichletBC or a list of DirichletBCs."
            dolfinx.fem.petsc.set_bc(b, bcs0[0], x0=x, alpha=alpha)

        ghost_update(b, PETSc.InsertMode.INSERT_VALUES, PETSc.ScatterMode.FORWARD)

else:

    def apply_lifting_and_set_bc(_b, _a, _bcs, _x, _alpha):  # type: ignore
        raise RuntimeError(
            "petsc4py is not available. Cannot apply lifting and set boundary conditions."
        )



from __future__ import annotations

from typing import Callable
import logging
from packaging.version import parse as _v

import numpy as np
import ufl
import dolfinx

__all__ = ["NewtonSolver", "BlockedNewtonSolver"]

if dolfinx.has_petsc4py and dolfinx.has_petsc:
    logger = logging.getLogger(__name__)
    from petsc4py import PETSc
    import dolfinx.fem.petsc

    # assemble_vector_block(scale=...) is renamed assemble_vector_block(alpha=...)
    # in 0.9
    _alpha_kw: str = "alpha"
    if _v(dolfinx.__version__) < _v("0.9"):
        _alpha_kw = "scale"

    class NewtonSolver:
        max_iterations: int
        _bcs: list[dolfinx.fem.DirichletBC]
        A: PETSc.Mat
        b: PETSc.Vec
        _J: dolfinx.fem.Form | ufl.form.Form
        _F: dolfinx.fem.Form | ufl.form.Form
        _dx: PETSc.Vec
        _error_on_convergence: bool
        _pre_solve_callback: Callable[["NewtonSolver"], None] | None
        _post_solve_callback: Callable[["NewtonSolver"], None] | None

        def __init__(
            self,
            F: list[dolfinx.fem.Form],
            J: list[list[dolfinx.fem.Form]],
            w: list[dolfinx.fem.Function],
            bcs: list[dolfinx.fem.DirichletBC] | None = None,
            max_iterations: int = 5,
            petsc_options: dict[str, str | float | int | None] | None = None,
            error_on_nonconvergence: bool = True,
        ):
            """
            Create a Newton solver for a block nonlinear problem ``F(u) = 0``.
            Solved as ``J(u) du = -F(u)``, where ``J`` is the Jacobian of ``F``.

            Args:
                F: List of forms defining the residual
                J: List of lists of forms defining the Jacobian
                w: List of functions representing the solution
                bcs: List of Dirichlet boundary conditions
                max_iterations: Maximum number of iterations in Newton solver
                petsc_options: PETSc options for Krylov subspace solver.
                error_on_nonconvergence: Raise an error if the linear solver
                    does not converge or Newton solver doesn't converge
            """
            # Compile forms if not compiled. Will throw error it requires entity maps
            self._F = dolfinx.fem.form(F)
            self._J = dolfinx.fem.form(J)

            # Store solution and accessible/modifiable properties
            self._pre_solve_callback = None
            self._post_solve_callback = None
            self.max_iterations = max_iterations
            self._error_on_convergence = error_on_nonconvergence
            self.bcs = [] if bcs is None else bcs
            self.w = w

            # Create PETSc objects for block assembly
            try:
                self.b = dolfinx.fem.petsc.create_vector_block(self._F)
                self.A = dolfinx.fem.petsc.create_matrix_block(self._J)
                self.dx = dolfinx.fem.petsc.create_vector_block(self._F)
                self.x = dolfinx.fem.petsc.create_vector_block(self._F)
            except AttributeError:
                self.b = dolfinx.fem.petsc.create_vector(
                    dolfinx.fem.extract_function_spaces(self._F), kind="mpi"
                )
                self.A = dolfinx.fem.petsc.create_matrix(self._J, kind="mpi")
                self.dx = dolfinx.fem.petsc.create_vector(
                    dolfinx.fem.extract_function_spaces(self._F), kind="mpi"
                )
                self.x = dolfinx.fem.petsc.create_vector(
                    dolfinx.fem.extract_function_spaces(self._F), kind="mpi"
                )

                self._maps = [
                    (
                        form.function_spaces[0].dofmaps(0).index_map,
                        form.function_spaces[0].dofmaps(0).index_map_bs,
                    )
                    for form in self.F
                ]

            # Set PETSc options
            opts = PETSc.Options()
            if petsc_options is not None:
                for k, v in petsc_options.items():
                    opts[k] = v

            # Define KSP solver
            self._solver = PETSc.KSP().create(self.b.getComm().tompi4py())
            self._solver.setOperators(self.A)
            self._solver.setFromOptions()

            # Set matrix and vector PETSc options
            self.A.setFromOptions()
            self.b.setFromOptions()

        def set_pre_solve_callback(self, callback: Callable[["NewtonSolver"], None]):
            """Set a callback function that is called before each Newton iteration."""
            self._pre_solve_callback = callback

        def set_post_solve_callback(self, callback: Callable[["NewtonSolver"], None]):
            """Set a callback function that is called after each Newton iteration."""
            self._post_solve_callback = callback

        def solve(self, atol=1e-6, rtol=1e-8, beta=1.0) -> int:
            """Solve the nonlinear problem using Newton's method.

            Args:
                atol: Absolute tolerance for the update.
                rtol: Relative tolerance for the update.
                beta: Damping parameter for the update.

            Returns:
                The number of Newton iterations used to converge.

            Note:
                The tolerance is on the 0-norm of the update.
            """
            i = 1

            while i <= self.max_iterations:
                if self._pre_solve_callback is not None:
                    self._pre_solve_callback(self)

                # Pack constants and coefficients
                try:
                    constants_L = dolfinx.fem.pack_constants(self._F)
                    coeffs_L = dolfinx.fem.pack_coefficients(self._F)
                    constants_a = [
                        dolfinx.fem.pack_constants(form)
                        if form is not None
                        else np.array([], dtype=self.x.array.dtype)
                        for form in self._J
                    ]
                    coeffs_a = [
                        {} if form is None else dolfinx.fem.pack_coefficients(form)
                        for form in self._J
                    ]
                except AttributeError:
                    # NOTE: DOLFINx 0.9 compatibility
                    # Pack constants and coefficients
                    constants_L = [
                        form and dolfinx.cpp.fem.pack_constants(form._cpp_object)
                        for form in self._F
                    ]
                    coeffs_L = [
                        dolfinx.cpp.fem.pack_coefficients(form._cpp_object) for form in self._F
                    ]
                    constants_a = [
                        [
                            dolfinx.cpp.fem.pack_constants(form._cpp_object)
                            if form is not None
                            else np.array([], dtype=self.x.array.dtype)
                            for form in forms
                        ]
                        for forms in self._J
                    ]
                    coeffs_a = [
                        [
                            {}
                            if form is None
                            else dolfinx.cpp.fem.pack_coefficients(form._cpp_object)
                            for form in forms
                        ]
                        for forms in self._J
                    ]
                # Scatter previous solution `w` to `self.x`, the blocked version used for lifting
                dolfinx.cpp.la.petsc.scatter_local_vectors(
                    self.x,
                    [si.x.petsc_vec.array_r for si in self.w],
                    [
                        (
                            si.function_space.dofmap.index_map,
                            si.function_space.dofmap.index_map_bs,
                        )
                        for si in self.w
                    ],
                )
                self.x.ghostUpdate(addv=PETSc.InsertMode.INSERT, mode=PETSc.ScatterMode.FORWARD)

                # Assemble F(u_{i-1}) - J(u_D - u_{i-1}) and set du|_bc= u_D - u_{i-1}
                with self.b.localForm() as b_local:
                    b_local.set(0.0)
                try:
                    dolfinx.fem.petsc.assemble_vector_block(
                        self.b,
                        self._F,
                        self._J,
                        bcs=self.bcs,
                        x0=self.x,
                        coeffs_a=coeffs_a,
                        constants_a=constants_a,
                        coeffs_L=coeffs_L,
                        constants_L=constants_L,
                        # dolfinx 0.8 compatibility
                        # this is called 'scale' in 0.8, 'alpha' in 0.9
                        **{_alpha_kw: -1.0},
                    )
                except AttributeError:
                    dolfinx.fem.petsc.assemble_vector(
                        self.b, self._F, coeffs=coeffs_L, constants=constants_L
                    )
                    bcs1 = dolfinx.fem.bcs_by_block(
                        dolfinx.fem.extract_function_spaces(self._J, 1), self.bcs
                    )
                    dolfinx.fem.petsc.apply_lifting(
                        self.b,
                        self._J,
                        bcs=bcs1,
                        x0=self.x,
                        coeffs=coeffs_a,
                        constants=constants_a,
                        alpha=-1.0,
                    )
                    self.b.ghostUpdate(addv=PETSc.InsertMode.ADD, mode=PETSc.ScatterMode.REVERSE)
                    bcs0 = dolfinx.fem.bcs_by_block(
                        dolfinx.fem.extract_function_spaces(self._F), self.bcs
                    )
                    dolfinx.fem.petsc.set_bc(self.b, bcs0, x0=self.x, alpha=-1.0)
                    self.b.ghostUpdate(PETSc.InsertMode.INSERT_VALUES, PETSc.ScatterMode.FORWARD)

                # Assemble Jacobian
                self.A.zeroEntries()
                try:
                    dolfinx.fem.petsc.assemble_matrix_block(
                        self.A, self._J, bcs=self.bcs, constants=constants_a, coeffs=coeffs_a
                    )
                except AttributeError:
                    dolfinx.fem.petsc.assemble_matrix(
                        self.A, self._J, self.bcs, coeffs=coeffs_a, constants=constants_a
                    )
                self.A.assemble()

                self._solver.solve(self.b, self.dx)
                if self._error_on_convergence:
                    if (status := self._solver.getConvergedReason()) <= 0:
                        raise RuntimeError(f"Linear solver did not converge, got reason: {status}")

                # Update solution
                offset_start = 0
                for s in self.w:
                    num_sub_dofs = (
                        s.function_space.dofmap.index_map.size_local
                        * s.function_space.dofmap.index_map_bs
                    )
                    s.x.petsc_vec.array_w[:num_sub_dofs] -= (
                        beta * self.dx.array_r[offset_start : offset_start + num_sub_dofs]
                    )
                    s.x.petsc_vec.ghostUpdate(
                        addv=PETSc.InsertMode.INSERT, mode=PETSc.ScatterMode.FORWARD
                    )
                    offset_start += num_sub_dofs

                if self._post_solve_callback is not None:
                    self._post_solve_callback(self)

                # Compute norm of update
                residual = self.dx.norm(PETSc.NormType.NORM_2)
                if i == 1:
                    self.residual_0 = residual
                relative_residual = residual / max(self.residual_0, atol)

                logger.info(
                    f"Newton iteration {i}"
                    f": r (abs) = {residual} (tol={atol}), "
                    f"r (rel) = {relative_residual} (tol={rtol})"
                )
                if relative_residual < rtol or residual < atol:
                    return i
                i += 1

            if self._error_on_convergence:
                raise RuntimeError("Newton solver did not converge")
            else:
                return self.max_iterations

        @property
        def F(self):
            """The list of residuals where each entry is a ``dolfinx.fem.Form``."""
            return self._F

        @property
        def J(self):
            """
            The Jacobian blocks represented as lists of lists where each entry
            is a ``dolfinx.fem.Form``.
            """
            return self._J

        def __del__(self):
            """Clean up the solver by destroying PETSc objects."""
            self.A.destroy()
            self.b.destroy()
            self.dx.destroy()
            self._solver.destroy()
            self.x.destroy()

    class BlockedNewtonSolver(dolfinx.cpp.nls.petsc.NewtonSolver):
        def __init__(
            self,
            F: list[ufl.form.Form],
            u: list[dolfinx.fem.Function],
            bcs: list[dolfinx.fem.DirichletBC] = [],
            J: list[list[ufl.form.Form]] | None = None,
            form_compiler_options: dict | None = None,
            jit_options: dict | None = None,
            petsc_options: dict | None = None,
            entity_maps: dict | None = None,
        ):
            """Initialize solver for solving a non-linear problem using Newton's method.
            Args:
                F: List of PDE residuals [F_0(u, v_0), F_1(u, v_1), ...]
                u: List of unknown functions u=[u_0, u_1, ...]
                bcs: List of Dirichlet boundary conditions
                J: UFL representation of the Jacobian (Optional)
                    Note:
                        If not provided, the Jacobian will be computed using the
                        assumption that the test functions come from a ``ufl.MixedFunctionSpace``
                form_compiler_options: Options used in FFCx
                    compilation of this form. Run ``ffcx --help`` at the
                    command line to see all available options.
                jit_options: Options used in CFFI JIT compilation of C
                    code generated by FFCx. See ``python/dolfinx/jit.py``
                    for all available options. Takes priority over all
                    other option values.
                petsc_options:
                    Options passed to the PETSc Krylov solver.
                entity_maps: Maps used to map entities between different meshes.
                    Only needed if the forms have not been compiled a priori,
                    and has coefficients, test, or trial functions that are defined on
                    different meshes.
            """
            # Initialize base class
            super().__init__(u[0].function_space.mesh.comm)

            # Set PETSc options for Krylov solver
            prefix = self.krylov_solver.getOptionsPrefix()
            if prefix is None:
                prefix = ""
            if petsc_options is not None:
                # Set PETSc options
                opts = PETSc.Options()
                opts.prefixPush(prefix)
                for k, v in petsc_options.items():
                    opts[k] = v
                opts.prefixPop()
                self.krylov_solver.setFromOptions()
            self._F = dolfinx.fem.form(
                F,
                form_compiler_options=form_compiler_options,
                jit_options=jit_options,
                entity_maps=entity_maps,
            )

            # Create the Jacobian matrix, dF/du
            if J is None:
                if _v(dolfinx.__version__) < _v("0.9"):
                    raise RuntimeError(
                        "Automatic computation of Jacobian for blocked problem is only"
                        + "supported in DOLFINx 0.9 and later"
                    )
                du = ufl.TrialFunctions(ufl.MixedFunctionSpace(*[ui.function_space for ui in u]))
                J = ufl.extract_blocks(
                    sum(ufl.derivative(sum(F), u[i], du[i]) for i in range(len(u)))
                )
            self._a = dolfinx.fem.form(
                J,
                form_compiler_options=form_compiler_options,
                jit_options=jit_options,
                entity_maps=entity_maps,
            )

            self._bcs = bcs
            self._u = u
            self._pre_solve_callback: Callable[["BlockedNewtonSolver"], None] | None = None
            self._post_solve_callback: Callable[["BlockedNewtonSolver"], None] | None = None

            # Create structures for holding arrays and matrix
            try:
                self._b = dolfinx.fem.petsc.create_vector_block(self._F)
                self._J = dolfinx.fem.petsc.create_matrix_block(self._a)
                self._dx = dolfinx.fem.petsc.create_vector_block(self._F)
                self._x = dolfinx.fem.petsc.create_vector_block(self._F)
            except AttributeError:
                self._b = dolfinx.fem.petsc.create_vector(
                    dolfinx.fem.extract_function_spaces(self._F), kind="mpi"
                )
                self._J = dolfinx.fem.petsc.create_matrix(self._a, kind="mpi")
                self._dx = dolfinx.fem.petsc.create_vector(
                    dolfinx.fem.extract_function_spaces(self._F), kind="mpi"
                )
                self._x = dolfinx.fem.petsc.create_vector(
                    dolfinx.fem.extract_function_spaces(self._F), kind="mpi"
                )

                self._maps = [
                    (
                        form.function_spaces[0].dofmaps(0).index_map,
                        form.function_spaces[0].dofmaps(0).index_map_bs,
                    )
                    for form in self._F
                ]

            self._J.setOptionsPrefix(prefix)
            self._J.setFromOptions()

            self.setJ(self._assemble_jacobian, self._J)
            self.setF(self._assemble_residual, self._b)
            self.set_form(self._pre_newton_iteration)
            self.set_update(self._update_function)

        def set_pre_solve_callback(self, callback: Callable[["BlockedNewtonSolver"], None]):
            """Set a callback function that is called before each Newton iteration."""
            self._pre_solve_callback = callback

        def set_post_solve_callback(self, callback: Callable[["BlockedNewtonSolver"], None]):
            """Set a callback function that is called after each Newton iteration."""
            self._post_solve_callback = callback

        @property
        def L(self) -> list[dolfinx.fem.Form]:
            """Compiled linear form (the residual form)"""
            return self._F

        @property
        def a(self) -> list[list[dolfinx.fem.Form]]:
            """Compiled bilinear form (the Jacobian form)"""
            return self._a

        @property
        def u(self):
            return self._u

        def __del__(self):
            self._J.destroy()
            self._b.destroy()
            self._dx.destroy()
            self._x.destroy()

        def _pre_newton_iteration(self, x: PETSc.Vec) -> None:
            """Function called before the residual or Jacobian is
            computed.
            Args:
            x: The vector containing the latest solution
            """
            if self._pre_solve_callback is not None:
                self._pre_solve_callback(self)
            # Scatter previous solution `u=[u_0, ..., u_N]` to `x`; the
            # blocked version used for lifting
            dolfinx.cpp.la.petsc.scatter_local_vectors(
                x,
                [ui.x.petsc_vec.array_r for ui in self._u],
                [
                    (
                        ui.function_space.dofmap.index_map,
                        ui.function_space.dofmap.index_map_bs,
                    )
                    for ui in self._u
                ],
            )
            x.ghostUpdate(addv=PETSc.InsertMode.INSERT, mode=PETSc.ScatterMode.FORWARD)

        def _assemble_residual(self, x: PETSc.Vec, b: PETSc.Vec) -> None:
            """Assemble the residual F into the vector b.
            Args:
                x: The vector containing the latest solution
                b: Vector to assemble the residual into
            """
            # Assemble F(u_{i-1}) - J(u_D - u_{i-1}) and set du|_bc= u_D - u_{i-1}
            with b.localForm() as b_local:
                b_local.set(0.0)
            try:
                dolfinx.fem.petsc.assemble_vector_block(
                    b,
                    self._F,
                    self._a,
                    bcs=self._bcs,
                    x0=x,
                    # dolfinx 0.8 compatibility
                    # this is called 'scale' in 0.8, 'alpha' in 0.9
                    **{_alpha_kw: -1.0},
                )
            except AttributeError:
                dolfinx.fem.petsc.assemble_vector(b, self._F)
                bcs1 = dolfinx.fem.bcs_by_block(
                    dolfinx.fem.extract_function_spaces(self._a, 1), self._bcs
                )
                dolfinx.fem.petsc.apply_lifting(b, self._a, bcs=bcs1, x0=x, alpha=-1.0)
                b.ghostUpdate(addv=PETSc.InsertMode.ADD, mode=PETSc.ScatterMode.REVERSE)
                bcs0 = dolfinx.fem.bcs_by_block(
                    dolfinx.fem.extract_function_spaces(self._F), self._bcs
                )
                dolfinx.fem.petsc.set_bc(b, bcs0, x0=x, alpha=-1.0)
            b.ghostUpdate(PETSc.InsertMode.INSERT_VALUES, PETSc.ScatterMode.FORWARD)

        def _assemble_jacobian(self, x: PETSc.Vec, A: PETSc.Mat) -> None:
            """Assemble the Jacobian matrix.
            Args:
                x: The vector containing the latest solution
            """
            # Assemble Jacobian
            A.zeroEntries()
            try:
                dolfinx.fem.petsc.assemble_matrix_block(A, self._a, bcs=self._bcs)
            except AttributeError:
                dolfinx.fem.petsc.assemble_matrix(A, self._a, self._bcs)
            A.assemble()

        def _update_function(self, solver, dx: PETSc.Vec, x: PETSc.Vec):
            if self._post_solve_callback is not None:
                self._post_solve_callback(self)
            # Update solution
            offset_start = 0
            for ui in self._u:
                Vi = ui.function_space
                num_sub_dofs = Vi.dofmap.index_map.size_local * Vi.dofmap.index_map_bs
                ui.x.petsc_vec.array_w[:num_sub_dofs] -= (
                    self.relaxation_parameter
                    * dx.array_r[offset_start : offset_start + num_sub_dofs]
                )
                ui.x.petsc_vec.ghostUpdate(
                    addv=PETSc.InsertMode.INSERT, mode=PETSc.ScatterMode.FORWARD
                )
                offset_start += num_sub_dofs

        def solve(self):
            """Solve non-linear problem into function. Returns the number
            of iterations and if the solver converged."""
            n, converged = super().solve(self._x)
            self._x.ghostUpdate(addv=PETSc.InsertMode.INSERT, mode=PETSc.ScatterMode.FORWARD)
            return n, converged

else:

    class NewtonSolver:  # type: ignore[no-redef]
        def __init__(self, *args, **kwargs):
            raise RuntimeError(
                "NewtonSolver is not available in this version of DOLFINx. "
                "Please install a version of DOLFINx with PETSc4py support."
            )

    class BlockedNewtonSolver:  # type: ignore[no-redef]
        def __init__(self, *args, **kwargs):
            raise RuntimeError(
                "BlockedNewtonSolver is not available in this version of DOLFINx. "
                "Please install a version of DOLFINx with PETSc4py support."
            )



# # Visualizing quadrature functions as point clouds
#
# Author: Henrik N.T. Finsberg
#
# SPDX-License-Identifier: MIT
#
# Quadrature functions are not possible to visualize directly in ParaView, as they are not defined on a mesh.
# However, we can visualize them as point clouds.
# In this example we will show how you can use `scifem` to save your quadrature fuctions as XDMF files,
# which can be loaded into ParaView for visualization.
#
# ```{note}
# This demo requires a backend for writing HDF5 files. Please install `scifem` with either `scifem[adios2]` or `scifem[h5py]`.
# ```

# First, we import the necessary modules.

import logging
from mpi4py import MPI
import numpy as np
import basix
import dolfinx
import scifem

# First initialize logging
logging.basicConfig(level=logging.INFO)

# Now let's create a quadrature function on a unit square mesh. We will use the `basix.ufl` module to create the quadrature element.

mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10, dolfinx.mesh.CellType.triangle, dtype=np.float64)
el = basix.ufl.quadrature_element(
    scheme="default", degree=3, cell=mesh.basix_cell(), value_shape=()
)
V = dolfinx.fem.functionspace(mesh, el)
u = dolfinx.fem.Function(V)
u.interpolate(lambda x: np.sin(np.pi * x[0]) * np.sin(np.pi * x[1]))


# One option that already exists with the current DOLFINx/Pyvista API is to plot them as points


import pyvista

plotter = pyvista.Plotter()
plotter.add_points(
    V.tabulate_dof_coordinates(),
    scalars=u.x.array,
    render_points_as_spheres=True,
    point_size=20,
    show_scalar_bar=False,
)
if not pyvista.OFF_SCREEN:
    plotter.show()


# Using `scifem`, we can write the point cloud data to an XDMFFile that can be opened with Paraview.

with scifem.xdmf.XDMFFile("point_cloud.xdmf", [u]) as xdmf:
    xdmf.write(0.0)

# The point cloud can now be loaded into ParaView for visualization, by selecting "Point Gaussian" as the representation.
# ![Point cloud in ParaView](../docs/_static/point_cloud.png)

# We can write any `dolfinx.fem.FunctionSpace` that has support for `tabulate_dof_coordinates` to a point cloud.
# For example, we can create a higher order Lagrange space, and write two functions to file.

Q = dolfinx.fem.functionspace(mesh, ("Lagrange", 3, (2, )))
q_1 = dolfinx.fem.Function(Q, name="q_sin")
q_1.interpolate(lambda x: (np.sin(np.pi * x[0]),  np.sin(np.pi * x[1])))
q_2 = dolfinx.fem.Function(Q, name="q_cos")
q_2.interpolate(lambda x: (np.cos(np.pi * x[0]),  np.cos(np.pi * x[1])))

# We write these two functions to file as illustrated above

with scifem.xdmf.XDMFFile("point_cloud_lagrange.xdmf", [q_1, q_2]) as xdmf:
    xdmf.write(0.0)

# yielding the following point clouds in ParaView after applying glyphs.
# ![Point cloud in ParaView](../docs/_static/cos_sin_pointcloud.png)

# If you have time dependent data, you can write multiple time steps to the same file using e.g

with scifem.xdmf.XDMFFile("point_cloud_lagrange.xdmf", [q_1, q_2]) as xdmf:
    xdmf.write(0.0)
    q_1.interpolate(lambda x: (-x[0], x[1]))
    q_2.interpolate(lambda x: (x[0], -x[1]))
    xdmf.write(0.3)

# If you need to keep the file open for longer, you can use the following syntax

xdmf = scifem.xdmf.XDMFFile("point_cloud_lagrange.xdmf", [q_1, q_2])
xdmf.write(0.0)
q_1.interpolate(lambda x: (-x[0], x[1]))
q_2.interpolate(lambda x: (x[0], -x[1]))
xdmf.write(0.3)
xdmf.close()



# # Transfer meshtags to submeshes
# Author: Jørgen S. Dokken
#
# SPDX-License-Identifier: MIT
#
# In DOLFINx, one can create sub-meshes of entities of any co-dimension $(0,\dots,\mathrm{tdim})$.
# For complex meshes, it is not easy to locate subdomains or boundaries of interest.
# In this example, we will import a mesh from GMSH, where we have marked different parts of the domain
# with different markers, and transfer these markers to a submesh.


import os
import sys
import pyvista
from scifem import transfer_meshtags_to_submesh
from mpi4py import MPI
import gmsh
import dolfinx
try:
    from dolfinx.io import gmsh as gmshio
except ImportError:
    import dolfinx.io.gmshio as gmshio

# We start by embedding an ellipsoid within another ellipsoid using GMSH.
# For more details about creating this mesh in GMSH,
# see: [FEniCS Workshop: External meshes](https://jsdokken.com/FEniCS-workshop/src/external_mesh.html)

gmsh.initialize()

center = (0, 0, 0)
aspect_ratio = 0.5
R_i = 0.3
R_e = 0.8

inner_disk = gmsh.model.occ.addDisk(*center, R_i, aspect_ratio * R_i)
outer_disk = gmsh.model.occ.addDisk(*center, R_e, R_e)
whole_domain, map_to_input = gmsh.model.occ.fragment(
    [(2, outer_disk)], [(2, inner_disk)])
gmsh.model.occ.synchronize()
circle_inner = [idx for (dim, idx) in map_to_input[1] if dim == 2]
circle_outer = [idx for (dim, idx) in map_to_input[0]
                if dim == 2 and idx not in circle_inner]
gmsh.model.addPhysicalGroup(2, circle_inner, tag=3)
gmsh.model.addPhysicalGroup(2, circle_outer, tag=7)
inner_boundary = gmsh.model.getBoundary(
    [(2, e) for e in circle_inner], recursive=False, oriented=False)
outer_boundary = gmsh.model.getBoundary(
    [(2, e) for e in circle_outer], recursive=False, oriented=False)
interface = [idx for (dim, idx) in inner_boundary if dim == 1]
ext_boundary = [
    idx for (dim, idx) in outer_boundary if idx not in interface and dim == 1]

gmsh.model.addPhysicalGroup(1, interface, tag=12)
gmsh.model.addPhysicalGroup(1, ext_boundary, tag=15)

gmsh.option.setNumber("Mesh.CharacteristicLengthMax", 0.2)
gmsh.model.mesh.generate(2)
gmsh.model.mesh.setOrder(3)
gmsh.model.mesh.optimize("Netgen")

# Next, we read this mesh and corresponding markers into DOLFINx

mesh_data = gmshio.model_to_mesh(
gmsh.model, MPI.COMM_WORLD, 0, gdim=2)
if hasattr(mesh_data, "mesh"):
    circular_mesh = mesh_data.mesh
    cell_marker = mesh_data.cell_tags
    facet_marker = mesh_data.facet_tags
else:
    circular_mesh, cell_marker, facet_marker = mesh_data

# We visualize the mesh and its markers

# + tags=["hide-input"]


def plot_mesh(mesh: dolfinx.mesh.Mesh, values=None):
    """
    Given a DOLFINx mesh, create a `pyvista.UnstructuredGrid`,
    and plot it and the mesh nodes
    """
    plotter = pyvista.Plotter()
    V_linear = dolfinx.fem.functionspace(mesh, ("Lagrange", 1))
    linear_grid = pyvista.UnstructuredGrid(*dolfinx.plot.vtk_mesh(V_linear))
    if mesh.geometry.cmap.degree > 1:
        ugrid = pyvista.UnstructuredGrid(*dolfinx.plot.vtk_mesh(mesh))
        if values is not None:
            ugrid.cell_data["Marker"] = values
        plotter.add_mesh(ugrid, style="points", color="b", point_size=10)
        ugrid = ugrid.tessellate()
        plotter.add_mesh(ugrid, show_edges=False)
        plotter.add_mesh(linear_grid, style="wireframe", color="black")

    else:
        if values is not None:
            linear_grid.cell_data["Marker"] = values
        plotter.add_mesh(linear_grid, show_edges=True)
    plotter.show_axes()
    plotter.view_xy()
    plotter.show()


plot_mesh(circular_mesh, cell_marker.values)
# -

# Next, we create a submesh, only extracting the upper half of the mesh
tdim = circular_mesh.topology.dim
submesh, cell_map, vertex_map, node_map = dolfinx.mesh.create_submesh(
    circular_mesh, tdim, dolfinx.mesh.locate_entities(circular_mesh, tdim, lambda x: x[1] > 0))

# We transfer the cell markers to the submesh

sub_cell_marker, sub_cell_map = transfer_meshtags_to_submesh(
    cell_marker, submesh, vertex_map, cell_map)

# and visualize it

# + tags=["hide-input"]
plot_mesh(submesh, sub_cell_marker.values)
# -



# Create a point source for Poisson problem
# Author: Jørgen S. Dokken
# SPDX-License-Identifier: MIT

from __future__ import annotations

from packaging.version import Version
from mpi4py import MPI
from petsc4py import PETSc

import dolfinx
import dolfinx.fem.petsc
import numpy as np
import numpy.typing as npt
import ufl

from .utils import unroll_dofmap

__all__ = ["PointSource"]


class PointSource:
    """Class for defining a point source in a given function space."""

    def __init__(
        self,
        V: dolfinx.fem.FunctionSpace,
        points: npt.NDArray[np.float32] | npt.NDArray[np.float64],
        magnitude: np.floating | np.complexfloating = dolfinx.default_scalar_type(1),
        tol: np.floating | None = None,
    ) -> None:
        """Initialize a point source.

        Args:
            V: The function space the point source is defined in.
            points: The points where the point source is located.
                Input shape: ``(num_points, 3)``
            magnitude: The magnitudes of the point sources.
            tol: Tolerance for point location. If `None` 100 times machine epsilon is used.

        Note:
            Points should only be defined on one process. If they are sent in
            from multiple processes, multiple point sources will be created.

        Note:
            If the point source is outside the mesh, a ``ValueError`` will be raised.
        """
        self._function_space = V
        if V.dofmap.bs > 1 and dolfinx.__version__ == "0.8.0":
            raise NotImplementedError(
                "Block function spaces are not supported in dolfinx 0.8.0. Please upgrade dolfinx"
            )
        if points.ndim == 1:
            points = points.reshape(1, -3)
        elif points.ndim == 2 and points.shape[1] != 3:
            raise ValueError(f"Points should have shape (num_points, 3), got {points.shape}")

        self._input_points = points
        self._magnitude = magnitude
        # Initialize empty arrays
        self._points = np.empty((0, 3), dtype=points.dtype)
        self._cells = np.empty(0, dtype=np.int32)
        num_dofs = self._function_space.dofmap.dof_layout.num_dofs * self._function_space.dofmap.bs
        self._basis_values = np.empty(
            (0, num_dofs), dtype=self._function_space.mesh.geometry.x.dtype
        )
        self._tol = (
            float(1e2 * np.finfo(self._input_points.dtype).eps) if tol is None else float(tol)
        )
        self.recompute_sources()
        self.compute_cell_contributions()

    def recompute_sources(self, tol: float | None = None):
        """Recompute the what cells the point sources collide with.

        This function should be called if the mesh geometry has been modified.

        Args:
            tol: Tolerance for point location. If `None`, the tolerance from initialization is used.
        """

        # Determine what process owns a point and what cells it lies within
        mesh = self._function_space.mesh
        tol = self._tol if tol is None else tol
        if dolfinx.__version__ == "0.8.0":
            src_ranks, _, self._points, self._cells = (
                dolfinx.cpp.geometry.determine_point_ownership(
                    mesh._cpp_object, self._input_points, tol
                )
            )
            self._points = np.array(self._points).reshape(-1, 3)
        elif Version(dolfinx.__version__) >= Version("0.9.0.0"):
            try:
                collision_data = dolfinx.cpp.geometry.determine_point_ownership(
                    mesh._cpp_object, self._input_points, tol
                )
            except TypeError:
                collision_data = dolfinx.geometry.determine_point_ownership(
                    mesh, self._input_points, padding=tol
                )

            self._points = collision_data.dest_points
            self._cells = collision_data.dest_cells
            src_ranks = collision_data.src_owner
        else:
            raise NotImplementedError(f"Unsupported version of dolfinx: {dolfinx.__version__}")
        if -1 in src_ranks:
            raise ValueError("Point source is outside the mesh.")

    def compute_cell_contributions(self):
        """Compute the basis function values at the point sources."""
        mesh = self._function_space.mesh
        # Pull owning points back to reference cell
        mesh_nodes = mesh.geometry.x
        cmap = mesh.geometry.cmap

        ref_x = np.zeros((len(self._cells), mesh.topology.dim), dtype=mesh.geometry.x.dtype)
        for i, (point, cell) in enumerate(zip(self._points, self._cells)):
            geom_dofs = mesh.geometry.dofmap[cell]
            ref_x[i] = cmap.pull_back(point.reshape(-1, 3), mesh_nodes[geom_dofs])

        # Create expression evaluating a trial function (i.e. just the basis function)
        u = ufl.TestFunction(self._function_space)
        bs = self._function_space.dofmap.bs
        num_dofs = self._function_space.dofmap.dof_layout.num_dofs * bs
        if len(self._cells) > 0:
            # NOTE: Expression lives on only this communicator rank
            # Expression is evaluated for every point in every cell, which means that we
            # need to discard values that are not on the diagonal.
            expr = dolfinx.fem.Expression(u, ref_x, comm=MPI.COMM_SELF)
            all_values = expr.eval(mesh, self._cells)
            # Diagonalize values (num_cells, num_points, num_dofs, bs) -> (num_cells, num_dofs)
            if Version(dolfinx.__version__) <= Version("0.9.0"):
                basis_values = np.empty((len(self._cells), num_dofs), dtype=all_values.dtype)
                for i in range(len(self._cells)):
                    basis_values[i] = sum(
                        [
                            all_values[i, i * num_dofs * bs : (i + 1) * num_dofs * bs][
                                j * num_dofs : (j + 1) * num_dofs
                            ]
                            for j in range(bs)
                        ]
                    )
            else:
                basis_values = np.zeros(
                    (len(self._cells), num_dofs), dtype=dolfinx.default_scalar_type
                )
                if bs == 1:
                    # Values have shape (num_cells, num_points, num_dofs)
                    for i in range(len(self._cells)):
                        basis_values[i] = all_values[i, i, :]
                else:
                    # Values have shape (num_cells, num_points, bs, num_dofs*bs)
                    for i in range(len(self._cells)):
                        basis_values[i] = sum(all_values[i, i, j, :] for j in range(bs))
        else:
            basis_values = np.zeros((0, num_dofs), dtype=dolfinx.default_scalar_type)
        self._basis_values = basis_values

    def apply_to_vector(
        self, b: dolfinx.fem.Function | dolfinx.la.Vector | PETSc.Vec, recompute: bool = False
    ):
        """Apply the point sources to a vector.

        Args:
            b: The vector to apply the point sources to.
            recompute: If the point sources should be recomputed before applying.
                Recomputation should only be done if the mesh geometry has been modified.

        Note:
            The user is responsible for forward scattering of the vector after
            applying the point sources.

        Note:
            If a PETSc vector is passed in, one has to call ``b.assemble()`` prior to solving the
            linear system (post scattering).
        """
        if recompute:
            self.recompute_sources()
            self.compute_cell_contributions()

        # Apply the point sources to the vector
        _dofs = self._function_space.dofmap.list[self._cells]
        unrolled_dofs = unroll_dofmap(_dofs, self._function_space.dofmap.bs)
        for dofs, values in zip(unrolled_dofs, self._basis_values, strict=True):
            if isinstance(b, dolfinx.fem.Function):
                b.x.array[dofs] += values * self._magnitude
            elif isinstance(b, dolfinx.la.Vector):
                b.array[dofs] += values * self._magnitude
            elif isinstance(b, PETSc.Vec):
                b.setValuesLocal(
                    dofs,
                    values * self._magnitude,
                    addv=PETSc.InsertMode.ADD_VALUES,
                )



from __future__ import annotations

from mpi4py import MPI
import ufl
import numpy as np
import dolfinx
from typing import Literal
import numpy.typing as npt
import typing


def _extract_dtype(expr: ufl.form.Form | ufl.core.expr.Expr | dolfinx.fem.Form) -> npt.DTypeLike:
    """Given a ufl form, expression or a compiled DOLFINx form extract the data type that should
    be used in form compilation.

    Args:
        form: The form to extract the data type from.

    """
    try:
        return expr.dtype
    except AttributeError:
        try:
            geom_type = ufl.domain.extract_domains(expr)[0].ufl_cargo().geometry.x.dtype
        except AttributeError:
            # Legacy support
            geom_type = expr.ufl_domain().ufl_cargo().geometry.x.dtype
        coefficients = expr.coefficients()
        constants = expr.constants()
        if (not coefficients) and (not constants):
            return geom_type
        else:
            data_types: set[npt.DTypeLike] = set()
            for coefficient in coefficients:
                data_types.add(coefficient.dtype)
            for constant in constants:
                data_types.add(constant.dtype)

            if len(data_types) > 1:
                raise RuntimeError("All coefficients and constants must have the same data type")
            assert len(data_types) == 1
            return data_types.pop()


def assemble_scalar(
    J: ufl.form.Form | dolfinx.fem.Form,
    entity_maps: typing.Optional[dict[dolfinx.mesh.Mesh, npt.NDArray[np.int32]]] = None,
) -> np.floating | np.complexfloating:
    """Assemble a scalar form and gather result across processes

    Args:
        form: The form to assemble.
        entity_maps: Maps of entities on related submeshes to the domain used in `J`.

    Returns:
        The accumulated value of the assembled form.
    """
    dtype = _extract_dtype(J)
    compiled_form = dolfinx.fem.form(J, entity_maps=entity_maps, dtype=dtype)

    if (rank := compiled_form.rank) != 0:
        raise ValueError(f"Form must be a scalar form, got for of arity {rank}")
    local_result = dolfinx.fem.assemble_scalar(compiled_form)
    return compiled_form.mesh.comm.allreduce(local_result, op=MPI.SUM)


def norm(
    expr: ufl.core.expr.Expr,
    norm_type: Literal["L2", "H1", "H10"],
    entity_maps: typing.Optional[dict[dolfinx.mesh.Mesh, npt.NDArray[np.int32]]] = None,
) -> dolfinx.fem.Form:
    """
    Compile the norm of an UFL expression into a DOLFINx form.

    Args:
        expr: UFL expression
        norm_type: Type of norm
        entity_maps: Mapping for Constants and Coefficients within the expression that
            lives on a submesh.
    """
    if norm_type == "L2":
        form = ufl.inner(expr, expr) * ufl.dx
    elif norm_type == "H1":
        form = ufl.inner(expr, expr) * ufl.dx + ufl.inner(ufl.grad(expr), ufl.grad(expr)) * ufl.dx
    elif norm_type == "H10":
        form = ufl.inner(ufl.grad(expr), ufl.grad(expr)) * ufl.dx
    else:
        raise RuntimeError(f"Unexpected norm type: {norm_type}")
    dtype = _extract_dtype(form)
    return dolfinx.fem.form(form, entity_maps=entity_maps, dtype=dtype)



from __future__ import annotations

from . import _scifem  # type: ignore
import collections
import dolfinx
import typing
import numpy as np
import numpy.typing as npt
from typing import Protocol
from packaging.version import Version
import basix
import ufl

__all__ = [
    "create_entity_markers",
    "transfer_meshtags_to_submesh",
    "SubmeshData",
    "reverse_mark_entities",
    "extract_submesh",
    "find_interface",
    "compute_interface_data",
    "compute_subdomain_exterior_facets",
    "create_geometry_function_space",
    "move",
]

if typing.TYPE_CHECKING:
    TaggedEntities = (
        tuple[int, typing.Callable[[npt.NDArray[np.floating]], npt.NDArray[np.bool_]]]
        | tuple[int, typing.Callable[[npt.NDArray[np.floating]], npt.NDArray[np.bool_]], bool]
    )


class _EntityMap(Protocol):
    """Protocol for EntityMap-like objects."""

    sub_topology: dolfinx.mesh.Topology
    dim: int

    def sub_topology_to_topology(
        self, entities: npt.NDArray[np.int32], inverse: bool
    ) -> npt.NDArray[np.int32]:
        """Map entities between sub-topology and topology."""
        ...


def get_entity_map(entity_map: _EntityMap | npt.NDArray[np.int32]) -> npt.NDArray[np.int32]:
    """Get an entity map from the sub-topology to the topology.

    This function handles both the deprecated construction of an entity map as a numpy array
    and the newer `EntityMap` class from `dolfinx.mesh`.

    Args:
        entity_map: An `EntityMap` object or a numpy array representing the mapping.
    Returns:
        Mapped indices of entities.
    """
    try:
        sub_top = entity_map.sub_topology
        assert isinstance(sub_top, dolfinx.mesh.Topology)
        sub_map = sub_top.index_map(entity_map.dim)
        indices = np.arange(sub_map.size_local + sub_map.num_ghosts, dtype=np.int32)
        return entity_map.sub_topology_to_topology(indices, inverse=False)
    except AttributeError:
        return entity_map


def create_entity_markers(
    domain: dolfinx.mesh.Mesh,
    dim: int,
    entities_list: list[TaggedEntities],
) -> dolfinx.mesh.MeshTags:
    """Mark entities of specified dimension according to a geometrical marker function.

    Args:
        domain: A ``dolfinx.mesh.Mesh`` object
        dim: Dimension of the entities to mark
        entities_list: A list of tuples with the following elements:

            - ``index 0``: The tag to assign to the entities
            - ``index 1``: A function that takes a point and returns a boolean array
              indicating whether the point is inside the entity
            - ``index 2``: Optional, if True, the entities will be marked on the boundary

    Returns:
        A ``dolfinx.mesh.MeshTags`` object with the corresponding entities marked.
        If an entity satisfies multiple input marker functions,
        it is not deterministic what value the entity gets.

    Note:

    """
    # Create required connectivities
    domain.topology.create_entities(dim)
    domain.topology.create_connectivity(dim, 0)
    domain.topology.create_connectivity(dim, domain.topology.dim)

    # Create marker function
    e_map = domain.topology.index_map(dim)
    num_entities_local = e_map.size_local + e_map.num_ghosts
    markers = np.full(num_entities_local, -1, dtype=np.int32)

    locate_entities = (
        lambda on_boundary: dolfinx.mesh.locate_entities_boundary
        if on_boundary
        else dolfinx.mesh.locate_entities
    )

    # Concatenate and sort the arrays based on indices
    for tagged_entity in entities_list:
        on_boundary = False if len(tagged_entity) == 2 else tagged_entity[2]
        entities = locate_entities(on_boundary)(domain, dim, tagged_entity[1])
        markers[entities] = tagged_entity[0]

    facets = np.flatnonzero(markers != -1).astype(np.int32)
    return dolfinx.mesh.meshtags(domain, dim, facets, markers[facets])


def transfer_meshtags_to_submesh(
    entity_tag: dolfinx.mesh.MeshTags,
    submesh: dolfinx.mesh.Mesh,
    vertex_to_parent: _EntityMap | npt.NDArray[np.int32],
    cell_to_parent: _EntityMap | npt.NDArray[np.int32],
) -> tuple[dolfinx.mesh.MeshTags, npt.NDArray[np.int32]]:
    """
    Transfer a ``entity_tag`` from a parent mesh to a ``submesh``.

    Args:
        entity_tag: Tag to transfer
        submesh: Submesh to transfer tag to
        vertex_to_parent: Mapping from submesh vertices to parent mesh vertices
        cell_to_parent: Mapping from submesh cells to parent entities
    Returns:
        A tuple (submesh_tag, sub_to_parent_entity_map) where: ``submesh_tag``
        is the tag on the submesh and ``sub_to_parent_entity_map`` is a mapping
        from submesh entities in the tag to the corresponding entities in the parent.
    """
    dim = entity_tag.dim
    sub_tdim = submesh.topology.dim
    if dim > sub_tdim:
        raise RuntimeError(
            f"Cannot transfer meshtags of dimension {dim} to submesh with topological dimension"
        )

    submesh.topology.create_connectivity(sub_tdim, sub_tdim)
    submesh.topology.create_connectivity(entity_tag.dim, 0)
    submesh.topology.create_connectivity(sub_tdim, entity_tag.dim)
    entity_tag.topology.create_connectivity(dim, 0)
    entity_tag.topology.create_connectivity(dim, sub_tdim)

    v_to_p = get_entity_map(vertex_to_parent)
    c_to_p = get_entity_map(cell_to_parent)
    cpp_tag, sub_to_parent_entity_map = _scifem.transfer_meshtags_to_submesh_int32(
        entity_tag._cpp_object, submesh.topology._cpp_object, v_to_p, c_to_p
    )
    return dolfinx.mesh.MeshTags(cpp_tag), sub_to_parent_entity_map


def reverse_mark_entities(
    entity_map: dolfinx.common.IndexMap, entities: npt.NDArray[np.int32]
) -> npt.NDArray[np.int32]:
    """Communicate entities marked on a single process to all processes that ghosts or
    owns this entity.

    Args:
        entity_map: Index-map describing entity ownership
        entities: Local indices of entities to communicate
    Returns:
        Local indices marked on any process sharing this entity
    """
    comm_vec = dolfinx.la.vector(entity_map, dtype=np.int32)
    comm_vec.array[:] = 0
    comm_vec.array[entities] = 1
    comm_vec.scatter_reverse(dolfinx.la.InsertMode.add)
    comm_vec.scatter_forward()
    return np.flatnonzero(comm_vec.array).astype(np.int32)


SubmeshData = collections.namedtuple(
    "SubmeshData", ["domain", "cell_map", "vertex_map", "node_map", "cell_tag"]
)


def extract_submesh(
    mesh: dolfinx.mesh.Mesh, entity_tag: dolfinx.mesh.MeshTags, tags: typing.Sequence[int]
) -> SubmeshData:
    """Generate a sub-mesh from a subset of tagged entities in a meshtag object.

    Args:
        mesh: The mesh to extract the submesh from.
        entity_tag: MeshTags object containing marked entities.
        tags: What tags the marked entities used in the submesh should have.

    Returns:
        A tuple `(submesh, subcell_to_parent_entity, subvertex_to_parent_vertex,
        subnode_to_parent_node, entity_tag_on_submesh)`.
    """

    # Accumulate all entities, including ghosts, for the specfic set of tagged entities
    edim = entity_tag.dim
    mesh.topology.create_connectivity(edim, mesh.topology.dim)
    tags_as_arr = np.asarray(tags, dtype=entity_tag.values.dtype)
    all_tagged_indices = np.isin(entity_tag.values, tags_as_arr)
    entities = entity_tag.indices[all_tagged_indices]
    # Extract submesh
    submesh, cell_map, vertex_map, node_map = dolfinx.mesh.create_submesh(mesh, edim, entities)

    # Transfer cell markers
    new_et, _ = transfer_meshtags_to_submesh(entity_tag, submesh, vertex_map, cell_map)
    new_et.name = entity_tag.name
    return SubmeshData(submesh, cell_map, vertex_map, node_map, new_et)


def find_interface(
    cell_tags: dolfinx.mesh.MeshTags,
    id_0: tuple[int, ...],
    id_1: tuple[int, ...],
) -> npt.NDArray[np.int32]:
    """Given to sets of cells, find the facets that are shared between them.

    Args:
        cell_tags: MeshTags object marking cells.
        id_0: Tags to extract for domain 0
        id_1: Tags to extract for domain 1

    Returns:
        The facets shared between the two domains.
    """
    topology = dolfinx.mesh.Topology(cell_tags.topology)

    assert topology.dim == cell_tags.dim
    tdim = topology.dim
    cell_map = topology.index_map(tdim)

    # Find all cells on process that has cell with tag(s) id_0
    domain_0 = reverse_mark_entities(
        cell_map,
        cell_tags.indices[
            np.isin(cell_tags.values, np.asarray(id_0, dtype=cell_tags.values.dtype))
        ],
    )

    # Find all cells on process that has cell with tag(s) id_1
    domain_1 = reverse_mark_entities(
        cell_map,
        cell_tags.indices[
            np.isin(cell_tags.values, np.asarray(id_1, dtype=cell_tags.values.dtype))
        ],
    )

    # Find all facets connected to each domain
    topology.create_connectivity(tdim, tdim - 1)
    facet_map = topology.index_map(tdim - 1)

    local_facets0 = dolfinx.mesh.compute_incident_entities(topology, domain_0, tdim, tdim - 1)
    facets0 = reverse_mark_entities(facet_map, local_facets0)

    local_facets1 = dolfinx.mesh.compute_incident_entities(topology, domain_1, tdim, tdim - 1)
    facets1 = reverse_mark_entities(facet_map, local_facets1)

    # Compute intersecting facets
    interface_facets = np.intersect1d(facets0, facets1)

    reverse_mark_entities(facet_map, interface_facets)

    topology.create_connectivity(tdim - 1, tdim)
    f_to_c = topology.connectivity(tdim - 1, tdim)
    num_cells_per_facet = f_to_c.offsets[interface_facets + 1] - f_to_c.offsets[interface_facets]
    is_interface = interface_facets[num_cells_per_facet == 2]
    return is_interface


def compute_subdomain_exterior_facets(
    mesh: dolfinx.mesh.Mesh, ct: dolfinx.mesh.MeshTags, markers: typing.Sequence[int]
) -> npt.NDArray[np.int32]:
    """Find the the facets that are considered to be on the "exterior" boundary of a subdomain.

    The subdomain is defined as the collection of cells in ``ct`` that is marked with any of the
    ``markers``. The exterior boundary of the subdomain is defined as the collection of facets
    that are only connected to a single cell within the subdomain.

    Note:
        Ghosted facets are included in the resulting array.

    Args:
        mesh: Mesh to extract subdomains from
        ct: MeshTags object marking subdomains
        markers: The tags making up the "new" mesh
    Returns:
        The exterior facets
    """
    # Create submesh to find the exterior facet of subdomain
    sub_mesh, cell_map, _, _, _ = extract_submesh(
        mesh,
        ct,
        markers,
    )
    sub_mesh.topology.create_connectivity(sub_mesh.topology.dim - 1, sub_mesh.topology.dim)
    sub_facets = dolfinx.mesh.exterior_facet_indices(sub_mesh.topology)

    # Map exterior facet to (submesh_cell, local_facet_index) tuples
    try:
        integration_entities = dolfinx.fem.compute_integration_domains(
            dolfinx.fem.IntegralType.exterior_facet, sub_mesh.topology, sub_facets
        )
    except TypeError:
        integration_entities = dolfinx.fem.compute_integration_domains(
            dolfinx.fem.IntegralType.exterior_facet,
            sub_mesh.topology,
            sub_facets,
            sub_mesh.topology.dim - 1,
        )
    integration_entities = integration_entities.reshape(-1, 2)
    submap_array = get_entity_map(cell_map)
    integration_entities[:, 0] = submap_array[integration_entities[:, 0]]

    # Get cell to facet connectivity (parent mesh)
    mesh.topology.create_connectivity(mesh.topology.dim, mesh.topology.dim - 1)
    num_facets_per_cell = dolfinx.cpp.mesh.cell_num_entities(
        mesh.topology.cell_type, mesh.topology.dim - 1
    )
    c_to_f = mesh.topology.connectivity(mesh.topology.dim, mesh.topology.dim - 1).array.reshape(
        -1, num_facets_per_cell
    )
    # Map (parent_cell, local_facet_index) to facet index (local to process)
    parent_facets = c_to_f[integration_entities[:, 0], integration_entities[:, 1]]
    facet_map = mesh.topology.index_map(mesh.topology.dim - 1)
    # Accumulate ghost facets
    return reverse_mark_entities(facet_map, parent_facets)


def compute_interface_data(
    cell_tags: dolfinx.mesh.MeshTags,
    facet_indices: npt.NDArray[np.int32],
    include_ghosts: bool = False,
) -> npt.NDArray[np.int32]:
    """
    Compute interior facet integrals that are consistently ordered according to the `cell_tags`,
    such that the data `(cell0, facet_idx0, cell1, facet_idx1)` is ordered such that
    `cell_tags[cell0]`<`cell_tags[cell1]`, i.e the cell with the lowest cell marker
    is considered the "+" restriction".

    Args:
        cell_tags: MeshTags that must contain an integer marker for all cells adjacent
            to the `facet_indices`
        facet_indices: List of facets (local index) that are on the interface.
        include_ghosts: If `True` integration entities will include facets that are ghosts on
        the current process. This is for instance useful for interpolation on interior facets.
    Returns:
        The integration data.
    """
    # Future compatibilty check
    integration_args: tuple[int] | tuple
    if Version("0.10.0") <= Version(dolfinx.__version__):
        integration_args = ()
    else:
        fdim = cell_tags.dim - 1
        integration_args = (fdim,)
    if include_ghosts:
        if len(facet_indices) == 0:
            return np.empty((0, 4), dtype=np.int32)
        f_to_c = cell_tags.topology.connectivity(cell_tags.topology.dim - 1, cell_tags.topology.dim)
        c_to_f = cell_tags.topology.connectivity(cell_tags.topology.dim, cell_tags.topology.dim - 1)

        # Extract the cells connected to each facet.
        # Assumption is that there can only be two cells per facet, and should always be
        # two cells per facet.
        num_cells_per_facet = f_to_c.offsets[facet_indices + 1] - f_to_c.offsets[facet_indices]
        assert np.all(num_cells_per_facet == 2), "All facets must be interior facets."
        facet_pos = np.vstack([f_to_c.offsets[facet_indices], f_to_c.offsets[facet_indices] + 1]).T
        cells = f_to_c.array[facet_pos].flatten()
        # Extract facets connected to all cells
        # Assumption is that all cells have the same number of facets
        num_facets_per_cell = c_to_f.offsets[1:] - c_to_f.offsets[:-1]
        assert all(
            num_facets_per_cell[cells.flatten()] == num_facets_per_cell[cells.flatten()[0]]
        ), "Cells must have facets."
        facets = np.vstack(
            [
                c_to_f.array[c_to_f.offsets[cells.flatten()] + i]
                for i in range(num_facets_per_cell[cells.flatten()[0]])
            ]
        ).T
        # Repeat facet indices twice to be able to do vectorized match
        rep_fi = np.repeat(facet_indices, 2)
        indicator = facets == rep_fi[:, None]
        _row, local_pos = np.nonzero(indicator)
        assert np.unique(_row).shape[0] == len(_row)
        idata = np.vstack([cells, local_pos]).T.reshape(-1, 4)
    else:
        idata = dolfinx.cpp.fem.compute_integration_domains(
            dolfinx.fem.IntegralType.interior_facet,
            cell_tags.topology,
            facet_indices,
            *integration_args,
        )
    ordered_idata = idata.reshape(-1, 4).copy()
    switch = cell_tags.values[ordered_idata[:, 0]] > cell_tags.values[ordered_idata[:, 2]]
    if True in switch:
        ordered_idata[switch, :] = ordered_idata[switch][:, [2, 3, 0, 1]]
    return ordered_idata


def create_geometry_function_space(
    mesh: dolfinx.mesh.Mesh, N: int | None = None
) -> dolfinx.fem.FunctionSpace:
    """
    Reconstruct a vector space with N components using
    the geometry dofmap to ensure a 1-1 mapping between mesh nodes and DOFs.

    Args:
        mesh: The mesh to create the function space on.
        N: The number of components. If not provided the geometrical dimension is chosen

    """
    geom_imap = mesh.geometry.index_map()
    geom_dofmap = mesh.geometry.dofmap
    ufl_domain = mesh.ufl_domain()
    assert ufl_domain is not None
    sub_el = ufl_domain.ufl_coordinate_element().sub_elements[0]
    adj_list = dolfinx.cpp.graph.AdjacencyList_int32(geom_dofmap)

    value_shape: tuple[int, ...]
    if N is None:
        ufl_el = basix.ufl.blocked_element(sub_el, shape=(mesh.geometry.dim,))
        value_shape = (mesh.geometry.dim,)
        N = value_shape[0]
    elif N == 1:
        ufl_el = sub_el
        value_shape = ()
    else:
        ufl_el = basix.ufl.blocked_element(sub_el, shape=(N,))
        value_shape = (N,)

    if ufl_el.dtype == np.float32:
        _fe_constructor = dolfinx.cpp.fem.FiniteElement_float32
        _fem_constructor = dolfinx.cpp.fem.FunctionSpace_float32
    elif ufl_el.dtype == np.float64:
        _fe_constructor = dolfinx.cpp.fem.FiniteElement_float64
        _fem_constructor = dolfinx.cpp.fem.FunctionSpace_float64
    else:
        raise RuntimeError(f"Unsupported type {ufl_el.dtype}")
    try:
        cpp_el = _fe_constructor(ufl_el.basix_element._e, block_shape=value_shape, symmetric=False)
    except TypeError:
        cpp_el = _fe_constructor(ufl_el.basix_element._e, block_size=N, symmetric=False)
    dof_layout = dolfinx.cpp.fem.create_element_dof_layout(cpp_el, [])
    cpp_dofmap = dolfinx.cpp.fem.DofMap(dof_layout, geom_imap, N, adj_list, N)

    # Create function space
    try:
        cpp_space = _fem_constructor(mesh._cpp_object, cpp_el, cpp_dofmap)
    except TypeError:
        cpp_space = _fem_constructor(mesh._cpp_object, cpp_el, cpp_dofmap, value_shape=value_shape)

    return dolfinx.fem.FunctionSpace(mesh, ufl_el, cpp_space)


def move(
    mesh: dolfinx.mesh.Mesh,
    u: dolfinx.fem.Function
    | ufl.core.expr.Expr
    | typing.Callable[[npt.NDArray[np.floating]], npt.NDArray[np.inexact]],
):
    """
    Move the geometry nodes of a mesh given by the movement of a function u.

    Args:
        mesh: The mesh to move
        u: The displacement as a :py:class:dolfinx.fem.Function`,
            :py:class:`ufl.core.expr.Expr` or a lambda function.
    """

    V_geom = create_geometry_function_space(mesh)
    u_geom = dolfinx.fem.Function(V_geom, dtype=mesh.geometry.x.dtype)
    if isinstance(u, dolfinx.fem.Function):
        u_geom.interpolate(u)
    elif isinstance(u, ufl.core.expr.Expr):
        u_compiled = dolfinx.fem.Expression(u, V_geom.element.interpolation_points)
        u_geom.interpolate(u_compiled)
    else:
        u_geom.interpolate(u)
    mesh.geometry.x[:, : mesh.geometry.dim] += u_geom.x.array[:].reshape(-1, mesh.geometry.dim)



from __future__ import annotations

import numpy.typing as npt
import numpy as np


def unroll_dofmap(dofs: npt.NDArray[np.int32], bs: int) -> npt.NDArray[np.int32]:
    """
    Given a two-dimensional dofmap of size `(num_cells, num_dofs_per_cell)`
    Expand the dofmap by its block size such that the resulting array
    is of size `(num_cells, bs*num_dofs_per_cell)`
    """
    num_cells, num_dofs_per_cell = dofs.shape
    unrolled_dofmap = np.repeat(dofs, bs).reshape(num_cells, num_dofs_per_cell * bs) * bs
    unrolled_dofmap += np.tile(np.arange(bs), num_dofs_per_cell)
    return unrolled_dofmap
